from functools import reduce
from operator import or_
from pychess.Utils.const import (
    WHITE,
    BLACK,
    KING,
    PAWN,
    EMPTY,
    KNIGHT,
    ROOK,
    BISHOP,
    QUEEN,
    A1,
    A2,
    A3,
    A4,
    A5,
    A6,
    A7,
    A8,
    B2,
    C3,
    D4,
    E5,
    F6,
    G7,
    H8,
    B7,
    C6,
    D5,
    E4,
    F3,
    G2,
    H1,
    B1,
    B8,
    H2,
    H7,
    G3,
    G6,
    B3,
    B6,
    sliders,
)

from .bitboard import bitPosArray, iterBits, setBit


def RANK(cord):
    return cord >> 3


def FILE(cord):
    return cord & 7


# Evaluating constants

PAWN_VALUE = 100
KNIGHT_VALUE = 300
BISHOP_VALUE = 330
ROOK_VALUE = 500
QUEEN_VALUE = 900
KING_VALUE = 2000
HAWK_VALUE = 800
ELEPHANT_VALUE = 850
PIECE_VALUES = [
    0,
    PAWN_VALUE,
    KNIGHT_VALUE,
    BISHOP_VALUE,
    ROOK_VALUE,
    QUEEN_VALUE,
    KING_VALUE,
    HAWK_VALUE,
    ELEPHANT_VALUE,
]

ASEAN_PIECE_VALUES = (0, 100, 450, 300, 630, 180, 2000)
CRAZY_PIECE_VALUES = (0, 100, 200, 240, 240, 380, 2000)
ATOMIC_PIECE_VALUES = (0, 100, 90, 0, 220, 850, 2000)

# Maximum possible search depth. The hash structure only allows 8-bit depths.
MAXPLY = 10
# Maximum possible score. Mate in n ply is +/- (MATE_VALUE-n).
# The hash structure only allows signed 16-bit scores.
MATE_VALUE = MAXVAL = 32767
MATE_DEPTH = 255


def VALUE_AT_PLY(val, ply):
    """Return the value of scoring val a given number of plies into the future."""
    if val >= +32512:
        return val - ply
    if val <= -32512:
        return val + ply
    return val


# How many points does it give to have the piece standing i cords from the
# opponent king
pawnTScale = [0, 40, 20, 12, 9, 6, 4, 2, 1, 0]
bishopTScale = [0, 50, 25, 15, 7, 5, 3, 2, 2, 1]
knightTScale = [0, 100, 50, 35, 10, 3, 2, 2, 1, 1]
rookTScale = [0, 50, 40, 15, 5, 2, 1, 1, 1, 0]
queenTScale = [0, 100, 60, 20, 10, 7, 5, 4, 3, 2]

passedScores = ((0, 48, 48, 120, 144, 192, 240, 0), (0, 240, 192, 144, 120, 48, 48, 0))

# Penalties for one or more isolated pawns on a given file
isolani_normal = (-8, -10, -12, -14, -14, -12, -10, -8)
# Penalties if the file is half-open (i.e. no enemy pawns on it)
isolani_weaker = (-22, -24, -26, -28, -28, -26, -24, -22)

# Distance boards for different pieces

taxicab = [[0] * 64 for i in range(64)]
sdistance = [[0] * 64 for i in range(64)]
for fcord in range(64):
    for tcord in range(fcord + 1, 64):
        fx = FILE(fcord)
        fy = RANK(fcord)
        tx = FILE(tcord)
        ty = RANK(tcord)
        taxicab[fcord][tcord] = taxicab[fcord][tcord] = abs(fx - tx) + abs(fy - ty)
        sdistance[fcord][tcord] = sdistance[fcord][tcord] = min(
            abs(fx - tx), abs(fy - ty)
        )

distance = [[[0] * 64 for i in range(64)] for j in range(KING + 1)]

distance[EMPTY] = None
distance[KING] = sdistance
distance[PAWN] = sdistance

# Special table for knightdistances

knightDistance = [
    6,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    6,
    5,
    4,
    5,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    5,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    3,
    2,
    3,
    4,
    1,
    2,
    1,
    4,
    3,
    2,
    3,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    1,
    2,
    3,
    2,
    1,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    3,
    2,
    3,
    2,
    3,
    0,
    3,
    2,
    3,
    2,
    3,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    1,
    2,
    3,
    2,
    1,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    3,
    2,
    3,
    4,
    1,
    2,
    1,
    4,
    3,
    2,
    3,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    3,
    4,
    3,
    2,
    3,
    2,
    3,
    2,
    3,
    4,
    3,
    4,
    5,
    4,
    5,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    3,
    4,
    5,
    4,
    5,
    6,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    4,
    5,
    6,
]

# Calculate

for fcord in range(64):
    frank = RANK(fcord)
    ffile = FILE(fcord)

    for tcord in range(fcord + 1, 64):
        # Notice, that we skip fcord == tcord, as all fields are zero from
        # scratch in anyway

        trank = RANK(tcord)
        tfile = FILE(tcord)

        # Knight
        field = (7 - frank + trank) * 15 + 7 - ffile + tfile
        distance[KNIGHT][tcord][fcord] = distance[KNIGHT][fcord][tcord] = (
            knightDistance[field]
        )

        # Rook
        if frank == trank or ffile == tfile:
            distance[ROOK][tcord][fcord] = distance[ROOK][fcord][tcord] = 1
        else:
            distance[ROOK][tcord][fcord] = distance[ROOK][fcord][tcord] = 2

        # Bishop
        if abs(frank - trank) == abs(ffile - tfile):
            distance[BISHOP][tcord][fcord] = distance[BISHOP][fcord][tcord] = 1
        else:
            distance[BISHOP][tcord][fcord] = distance[BISHOP][fcord][tcord] = 2

        # Queen
        if frank == trank or ffile == tfile or abs(frank - trank) == abs(ffile - tfile):
            distance[QUEEN][tcord][fcord] = distance[QUEEN][fcord][tcord] = 1
        else:
            distance[QUEEN][tcord][fcord] = distance[QUEEN][fcord][tcord] = 2

# Special cases for knights in corners
distance[KNIGHT][A1][B2] = distance[KNIGHT][B2][A1] = 4
distance[KNIGHT][H1][G2] = distance[KNIGHT][G2][H1] = 4
distance[KNIGHT][A8][B7] = distance[KNIGHT][B7][A8] = 4
distance[KNIGHT][H8][G7] = distance[KNIGHT][G7][H8] = 4

###############################################################################
# Boards used for evaluating
###############################################################################

pawnScoreBoard = (
    (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        5,
        5,
        5,
        -10,
        -10,
        5,
        5,
        5,
        -2,
        -2,
        -2,
        6,
        6,
        -2,
        -2,
        -2,
        0,
        0,
        0,
        25,
        25,
        0,
        0,
        0,
        2,
        2,
        12,
        16,
        16,
        12,
        2,
        2,
        4,
        8,
        12,
        16,
        16,
        12,
        4,
        4,
        4,
        8,
        12,
        16,
        16,
        12,
        4,
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ),
    (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        4,
        8,
        12,
        16,
        16,
        12,
        4,
        4,
        4,
        8,
        12,
        16,
        16,
        12,
        4,
        4,
        2,
        2,
        12,
        16,
        16,
        12,
        2,
        2,
        0,
        0,
        0,
        25,
        25,
        0,
        0,
        0,
        -2,
        -2,
        -2,
        6,
        6,
        -2,
        -2,
        -2,
        5,
        5,
        5,
        -10,
        -10,
        5,
        5,
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ),
)

outpost = (
    (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ),
    (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ),
)

normalKing = (
    24,
    24,
    24,
    16,
    16,
    0,
    32,
    32,
    24,
    20,
    16,
    12,
    12,
    16,
    20,
    24,
    16,
    12,
    8,
    4,
    4,
    8,
    12,
    16,
    12,
    8,
    4,
    0,
    0,
    4,
    8,
    12,
    12,
    8,
    4,
    0,
    0,
    4,
    8,
    12,
    16,
    12,
    8,
    4,
    4,
    8,
    12,
    16,
    24,
    20,
    16,
    12,
    12,
    16,
    20,
    24,
    24,
    24,
    24,
    16,
    16,
    0,
    32,
    32,
)

endingKing = (
    0,
    6,
    12,
    18,
    18,
    12,
    6,
    0,
    6,
    12,
    18,
    24,
    24,
    18,
    12,
    6,
    12,
    18,
    24,
    32,
    32,
    24,
    18,
    12,
    18,
    24,
    32,
    48,
    48,
    32,
    24,
    18,
    18,
    24,
    32,
    48,
    48,
    32,
    24,
    18,
    12,
    18,
    24,
    32,
    32,
    24,
    18,
    12,
    6,
    12,
    18,
    24,
    24,
    18,
    12,
    6,
    0,
    6,
    12,
    18,
    18,
    12,
    6,
    0,
)

racingKing = (
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    500,
    500,
    500,
    500,
    500,
    500,
    500,
    500,
    950,
    950,
    950,
    950,
    950,
    950,
    950,
    950,
    1500,
    1500,
    1500,
    1500,
    1500,
    1500,
    1500,
    1500,
    2500,
    2500,
    2500,
    2500,
    2500,
    2500,
    2500,
    2500,
    4000,
    4000,
    4000,
    4000,
    4000,
    4000,
    4000,
    4000,
    6000,
    6000,
    6000,
    6000,
    6000,
    6000,
    6000,
    6000,
    14000,
    14000,
    14000,
    14000,
    14000,
    14000,
    14000,
    14000,
)

# Maps for bitboards

d2e2 = (0x0018000000000000, 0x0000000000001800)
brank7 = (0x000000000000FF00, 0x00FF000000000000)
brank8 = (0x00000000000000FF, 0xFF00000000000000)
brank67 = (0x0000000000FFFF00, 0x00FFFF0000000000)
brank58 = (0x00000000FFFFFFFF, 0xFFFFFFFF00000000)
brank48 = (0x000000FFFFFFFFFF, 0xFFFFFFFFFF000000)

# Penalties if the file is half-open (i.e. no enemy pawns on it)
isolani_weaker = (-22, -24, -26, -28, -28, -26, -24, -22)

stonewall = [0, 0]
# D4, E3, F4
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - # - # - -
# - - - - # - - -
# - - - - - - - -
# - - - - - - - -
stonewall[WHITE] = 0x81400000000

# D5, E6, F5
# - - - - - - - -
# - - - - - - - -
# - - - - # - - -
# - - - # - # - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
stonewall[BLACK] = 0x81400000000

# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - # - - - - # -
# # # - - - - # #
# - - - - - - - -
qwingpawns1 = (bitPosArray[A2] | bitPosArray[B2], bitPosArray[A7] | bitPosArray[B7])
qwingpawns2 = (bitPosArray[A2] | bitPosArray[B3], bitPosArray[A7] | bitPosArray[B6])
kwingpawns1 = (bitPosArray[G2] | bitPosArray[H2], bitPosArray[G7] | bitPosArray[H7])
kwingpawns2 = (bitPosArray[G3] | bitPosArray[H2], bitPosArray[G6] | bitPosArray[H7])

################################################################################
#  Ranks and files                                                             #
################################################################################

rankBits = [255 << i * 8 for i in range(7, -1, -1)]
fileBits = [0x0101010101010101 << i for i in range(7, -1, -1)]

# Bit boards

WHITE_SQUARES = 0x55AA55AA55AA55AA
BLACK_SQUARES = 0xAA55AA55AA55AA55

# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
# - - - # # - - -
# - - - # # - - -
# - - - - - - - -
# - - - - - - - -
# - - - - - - - -
CENTER_FOUR = 0x0000001818000000

# - - - - - - - -
# - - - - - - - -
# - - # # # # - -
# - - # # # # - -
# - - # # # # - -
# - - # # # # - -
# - - - - - - - -
# - - - - - - - -
sbox = 0x00003C3C3C3C0000

# - - - - - - - -
# - # # # # # # -
# - # # # # # # -
# - # # # # # # -
# - # # # # # # -
# - # # # # # # -
# - # # # # # # -
# - - - - - - - -
lbox = 0x007E7E7E7E7E7E00

# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
# - - - - - # # #
right = fileBits[5] | fileBits[6] | fileBits[7]

# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
# # # # - - - - -
left = fileBits[0] | fileBits[1] | fileBits[2]

#  Generate the move bitboards.  For e.g. the bitboard for all                 #
#  the moves of a knight on f3 is given by MoveArray[knight][21].              #

dir = [
    None,
    [9, 11],  # Only capture moves are included
    [-21, -19, -12, -8, 8, 12, 19, 21],
    [-11, -9, 9, 11],
    [-10, -1, 1, 10],
    [-11, -10, -9, -1, 1, 9, 10, 11],
    [-11, -10, -9, -1, 1, 9, 10, 11],
    [-11, -9, 9, 11],
    [-10, -1, 1, 10],
    [-9, -11],
    [-11, -9, 9, 10, 11],
    [-11, -10, -9, 9, 11],
    [-11, -9, 9, 11],
    # Following are for front and back walls. Will be removed from list after
    # the loop
    [9, 10, 11],
    [-9, -10, -11],
]

sliders += [False, False]

map = [
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    -1,
    -1,
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    15,
    -1,
    -1,
    16,
    17,
    18,
    19,
    20,
    21,
    22,
    23,
    -1,
    -1,
    24,
    25,
    26,
    27,
    28,
    29,
    30,
    31,
    -1,
    -1,
    32,
    33,
    34,
    35,
    36,
    37,
    38,
    39,
    -1,
    -1,
    40,
    41,
    42,
    43,
    44,
    45,
    46,
    47,
    -1,
    -1,
    48,
    49,
    50,
    51,
    52,
    53,
    54,
    55,
    -1,
    -1,
    56,
    57,
    58,
    59,
    60,
    61,
    62,
    63,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
    -1,
]

moveArray = [[0] * 64 for i in range(len(dir))]  # moveArray[len(dir)][64]

for piece in range(1, len(dir)):
    for fcord in range(120):
        f = map[fcord]
        if f == -1:
            # We only generate moves for squares inside the board
            continue
        # Create a new bitboard
        b = 0
        for d in dir[piece]:
            tcord = fcord
            while True:
                tcord += d
                t = map[tcord]
                if t == -1:
                    # If we landed outside of board, there is no more to look
                    # for
                    break
                b = setBit(b, t)
                if not sliders[piece]:
                    # If we are a slider, we should not break, but add the dir
                    # value once again
                    break
        moveArray[piece][f] = b

frontWall = (moveArray[-2], moveArray[-1])
del moveArray[-1]
del moveArray[-1]
del dir[-1]
del dir[-1]
del sliders[-1]
del sliders[-1]

# For each square, there are 8 rays.  The first 4 rays are diagonals
# for the bishops and the next 4  are file/rank for the rooks.
# The queen uses all 8 rays.
# These rays are used for move generation rather than MoveArray[].
# Also initialize the directions[][] array.  directions[f][t] returns
# the index into rays[f] array allow us to find the ray in that direction.

directions = [[-1] * 64 for i in range(64)]  # directions[64][64]
rays = [[0] * 8 for i in range(64)]  # rays[64][8]

for fcord in range(120):
    f = map[fcord]
    if f == -1:
        continue
    ray = -1
    for piece in BISHOP, ROOK:
        for d in dir[piece]:
            ray += 1
            b = 0
            tcord = fcord
            while True:
                tcord += d
                t = map[tcord]
                if t == -1:
                    break
                rays[f][ray] = setBit(rays[f][ray], t)
                directions[f][t] = ray

# The FromToRay[b2][f6] gives the diagonal ray from c3 to f6;
# It also produces horizontal/vertical rays as well. If no
# ray is possible, then a 0 is returned.

fromToRay = [[0] * 64 for i in range(64)]  # fromToRay[64][64]

for piece in BISHOP, ROOK:
    for fcord in range(120):
        f = map[fcord]
        if f == -1:
            continue
        for d in dir[piece]:
            tcord = fcord
            t = map[tcord]

            while True:
                b = fromToRay[f][t]
                tcord += d
                t = map[tcord]
                if t == -1:
                    break
                fromToRay[f][t] = setBit(fromToRay[f][t], t)
                fromToRay[f][t] |= b

# The PassedPawnMask variable is used to determine if a pawn is passed.
#  his mask is basically all 1's from the square in front of the pawn to
# the promotion square, also duplicated on both files besides the pawn
# file. Other bits will be set to zero.
# E.g. PassedPawnMask[white][b3] = 1's in a4-c4-c8-a8 rect, 0 otherwise.

passedPawnMask = [[0] * 64, [0] * 64]

#  Do for white pawns first
for cord in range(64):
    passedPawnMask[WHITE][cord] = rays[cord][7]
    passedPawnMask[BLACK][cord] = rays[cord][4]
    if cord & 7 != 0:
        #  If file is not left most
        passedPawnMask[WHITE][cord] |= rays[cord - 1][7]
        passedPawnMask[BLACK][cord] |= rays[cord - 1][4]
    if cord & 7 != 7:
        #  If file is not right most
        passedPawnMask[WHITE][cord] |= rays[cord + 1][7]
        passedPawnMask[BLACK][cord] |= rays[cord + 1][4]

# The IsolaniMask variable is used to determine if a pawn is an isolani.
# This mask is basically all 1's on files beside the file the pawn is on.
# Other bits will be set to zero.
# E.g. isolaniMask[d-file] = 1's in c-file & e-file, 0 otherwise.

isolaniMask = [0] * 8

isolaniMask[0] = fileBits[1]
isolaniMask[7] = fileBits[6]
for i in range(1, 7):
    isolaniMask[i] = fileBits[i - 1] | fileBits[i + 1]

# The SquarePawnMask is used to determine if a king is in the square of
# the passed pawn and is able to prevent it from queening.
# Caveat:  Pawns on 2nd rank have the same mask as pawns on the 3rd rank
# as they can advance 2 squares.

squarePawnMask = [[0] * 64, [0] * 64]
for cord in range(64):
    # White mask
    rank = 7 - RANK(cord)
    i = max(cord & 56, cord - rank)
    j = min(cord | 7, cord + rank)
    for k in range(i, j + 1):
        squarePawnMask[WHITE][cord] |= bitPosArray[k] | fromToRay[k][k | 56]

    # Black mask
    rank = RANK(cord)
    i = max(cord & 56, cord - rank)
    j = min(cord | 7, cord + rank)
    for k in range(i, j + 1):
        squarePawnMask[BLACK][cord] |= bitPosArray[k] | fromToRay[k][k & 7]

# For pawns on 2nd rank, they have same mask as pawns on 3rd rank
for cord in range(A2, H2 + 1):
    squarePawnMask[WHITE][cord] = squarePawnMask[WHITE][cord + 8]
for cord in range(A7, H7 + 1):
    squarePawnMask[BLACK][cord] = squarePawnMask[BLACK][cord - 8]

# These tables are used to calculate rook, queen and bishop moves

ray00 = [rays[cord][5] | rays[cord][6] | 1 << (63 - cord) for cord in range(64)]
ray45 = [rays[cord][0] | rays[cord][3] | 1 << (63 - cord) for cord in range(64)]
ray90 = [rays[cord][4] | rays[cord][7] | 1 << (63 - cord) for cord in range(64)]
ray135 = [rays[cord][1] | rays[cord][2] | 1 << (63 - cord) for cord in range(64)]

attack00 = [{} for a in range(64)]
attack45 = [{} for a in range(64)]
attack90 = [{} for a in range(64)]
attack135 = [{} for a in range(64)]

cmap = [128, 64, 32, 16, 8, 4, 2, 1]
rot1 = [A1, A2, A3, A4, A5, A6, A7, A8]
rot2 = [A1, B2, C3, D4, E5, F6, G7, H8]
rot3 = [A8, B7, C6, D5, E4, F3, G2, H1]

# To save time, we init a main line for each of the four directions, and next
# we will translate it for each possible cord
for cord in range(8):
    for map in range(1, 256):
        # Skip entries without cord set, as cord will always be set
        if not map & cmap[cord]:
            continue

        # Find limits inclusive
        cord1 = cord2 = cord
        while cord1 > 0:
            cord1 -= 1
            if cmap[cord1] & map:
                break
        while cord2 < 7:
            cord2 += 1
            if cmap[cord2] & map:
                break

        # Remember A1 is the left most bit
        map00 = map << 56

        attack00[cord][map00] = fromToRay[cord][cord1] | fromToRay[cord][cord2]

        map90 = reduce(or_, (1 << 63 - rot1[c] for c in iterBits(map00)))
        attack90[rot1[cord]][map90] = (
            fromToRay[rot1[cord]][rot1[cord1]] | fromToRay[rot1[cord]][rot1[cord2]]
        )

        map45 = reduce(or_, (1 << 63 - rot2[c] for c in iterBits(map00)))
        attack45[rot2[cord]][map45] = (
            fromToRay[rot2[cord]][rot2[cord1]] | fromToRay[rot2[cord]][rot2[cord2]]
        )

        map135 = reduce(or_, (1 << 63 - rot3[c] for c in iterBits(map00)))
        attack135[rot3[cord]][map135] = (
            fromToRay[rot3[cord]][rot3[cord1]] | fromToRay[rot3[cord]][rot3[cord2]]
        )

MAXBITBOARD = (1 << 64) - 1

for r in range(A2, A8 + 1, 8):
    for cord in iterBits(ray00[r]):
        attack00[cord] = {map >> 8: ray >> 8 for map, ray in attack00[cord - 8].items()}

for r in range(B1, H1 + 1):
    for cord in iterBits(ray90[r]):
        attack90[cord] = {map >> 1: ray >> 1 for map, ray in attack90[cord - 1].items()}

# Bottom right
for r in range(B1, H1 + 1):
    for cord in iterBits(ray45[r]):
        attack45[cord] = {
            map << 8 & MAXBITBOARD: ray << 8 & MAXBITBOARD
            for map, ray in attack45[cord + 8].items()
        }

# Top left
for r in reversed(range(A8, H8)):
    for cord in iterBits(ray45[r]):
        attack45[cord] = {map >> 8: ray >> 8 for map, ray in attack45[cord - 8].items()}

# Top right
for r in range(B8, H8 + 1):
    for cord in iterBits(ray135[r]):
        attack135[cord] = {
            map >> 8: ray >> 8 for map, ray in attack135[cord - 8].items()
        }

# Bottom left
for r in reversed(range(A1, H1)):
    for cord in iterBits(ray135[r]):
        attack135[cord] = {
            map << 8 & MAXBITBOARD: ray << 8 & MAXBITBOARD
            for map, ray in attack135[cord + 8].items()
        }



import datetime

from gi.repository import GObject
from .BoardManager import names, months, dates

from pychess.ic import (
    GAME_TYPES_BY_SHORT_FICS_NAME,
    BLKCMD_STORED,
    BLKCMD_HISTORY,
    BLKCMD_JOURNAL,
)
from pychess.ic.FICSObjects import FICSAdjournedGame, FICSHistoryGame, FICSJournalGame

from pychess.Utils.const import (
    WON_ADJUDICATION,
    DRAW_AGREE,
    WON_DISCONNECTION,
    WON_CALLFLAG,
    WON_MATE,
    DRAW_INSUFFICIENT,
    DRAW_REPETITION,
    WON_RESIGN,
    DRAW_STALEMATE,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    WON_NOMATERIAL,
    DRAW_50MOVES,
    WHITEWON,
    DRAW,
    BLACK,
    WHITE,
    BLACKWON,
    reprResult,
    ADJOURNED,
)

from pychess.System.Log import log

reasons_dict = {
    "Adj": WON_ADJUDICATION,
    "Agr": DRAW_AGREE,
    "Dis": WON_DISCONNECTION,
    "Fla": WON_CALLFLAG,
    "Mat": WON_MATE,
    "NM": DRAW_INSUFFICIENT,
    "Rep": DRAW_REPETITION,
    "Res": WON_RESIGN,
    "Sta": DRAW_STALEMATE,
    "TM": DRAW_BLACKINSUFFICIENTANDWHITETIME,  # DRAW_WHITEINSUFFICIENTANDBLACKTIME
    "WLM": WON_NOMATERIAL,
    "WNM": WON_NOMATERIAL,
    "50": DRAW_50MOVES,
}

reasons = "(%s)" % "|".join(reasons_dict.keys())
ratings = r"([0-9\ \-\+]{1,4}[P E]?|UNR)"


class AdjournManager(GObject.GObject):
    __gsignals__ = {
        "adjournedGameAdded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onAdjournmentsList": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "historyGameAdded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onHistoryList": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "journalGameAdded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onJournalList": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)

        self.connection = connection

        self.connection.expect_line(
            self.__onStoredResponseNO, r"%s has no adjourned games\." % names
        )

        self.connection.expect_line(
            self.__onHistoryResponseNO, r"%s has no history games\." % names
        )

        self.connection.expect_line(
            self.__onJournalResponseNO,
            r"(%s has no journal entries\.)|(That journal is private.)" % names,
        )

        self.connection.expect_fromABplus(
            self.__onStoredResponseYES,
            "Stored games for %s:" % names,
            r"\s*C Opponent\s+On Type\s+Str\s+M\s+ECO\s+Date",
            r"\s*\d+: (B|W) %s\s+(Y|N) \[([a-z ]{3})\s+(\d+)\s+(\d+)\]\s+(\d+)-(\d+)\s+(W|B)(\d+)\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s"
            % (names, dates),
        )

        self.connection.expect_fromABplus(
            self.__onHistoryResponseYES,
            "History for %s:" % names,
            r"\s*Opponent\s+Type\s+ECO\s+End\s+Date",
            r"\s*(\d+): (-|\+|=)\s+(\d+)\s+(W|B)\s+(\d+) %s\s+\[([a-z ]{3})\s*(\d+)\s+(\d+)\]\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s\s+%s"
            % (names, reasons, dates),
        )

        self.connection.expect_fromABplus(
            self.__onJournalResponseYES,
            "Journal for %s:" % names,
            r"\s*White\s+Rating\s+Black\s+Rating\s+Type\s+ECO\s+End\s+Result",
            r"\s*%%(\d+): %s\s+%s\s+%s\s+%s\s+\[([a-z ]{3})\s+(\d+)\s+(\d+)\]\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s\s+(\*|1/2-1/2|1-0|0-1)"
            % (names, ratings, names, ratings, reasons),
        )

        self.connection.expect_line(
            self.__onAdjournedGameResigned, r"You have resigned the game\."
        )

        self.connection.bm.connect("curGameEnded", self.__onCurGameEnded)

        self.queryAdjournments()
        self.queryHistory()
        self.queryJournal()

        # TODO: Connect to {Game 67 (MAd vs. Sandstrom) Game adjourned by mutual agreement} *
        # TODO: Connect to adjourned game as adjudicated

    def __onStoredResponseYES(self, matchlist):
        # Stored games for User:
        #     C Opponent     On Type          Str  M    ECO Date
        #  1: W TheDane       N [ br  2  12]  0-0  B2   ??? Sun Nov 23,  6:14 CST 1997
        #  2: W PyChess       Y [psu  2  12] 39-39 W3   C20 Sun Jan 11, 17:40 ??? 2009
        #  3: B cjavad        N [ wr  2   2] 31-31 W18  --- Wed Dec 23, 06:58 PST 2009
        self.connection.stored_owner = matchlist[0].groups()[0]
        adjournments = []
        for match in matchlist[2:]:
            our_color = match.groups()[0]
            opponent_name, opponent_online = match.groups()[1:3]
            game_type = match.groups()[3]
            minutes, gain = match.groups()[4:6]
            str_white, str_black = match.groups()[6:8]
            next_color = match.groups()[8]
            move_num = match.groups()[9]
            week, month, day, hour, minute, timezone, year = match.groups()[11:18]
            gametime = datetime.datetime(
                int(year), months.index(month) + 1, int(day), int(hour), int(minute)
            )
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            our_color = our_color == "B" and BLACK or WHITE
            minutes = int(minutes)
            gain = int(gain)
            length = (int(move_num) - 1) * 2
            if next_color == "B":
                length += 1

            user = self.connection.players.get(self.connection.stored_owner)
            opponent = self.connection.players.get(opponent_name)
            wplayer, bplayer = (
                (user, opponent) if our_color == WHITE else (opponent, user)
            )
            game = FICSAdjournedGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                our_color=our_color,
                length=length,
                time=gametime,
                minutes=minutes,
                inc=gain,
                private=private,
            )
            if game.opponent.adjournment is False:
                game.opponent.adjournment = True

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("adjournedGameAdded", game)
            adjournments.append(game)

        self.emit("onAdjournmentsList", adjournments)

    __onStoredResponseYES.BLKCMD = BLKCMD_STORED

    def __onHistoryResponseYES(self, matchlist):
        # History for User:
        # Opponent      Type         ECO End Date
        # 66: - 1735 B    0 GuestHKZX     [ bu  3   0] B23 Res Sun Dec  6, 15:50 EST 2015
        # 67: - 1703 B    0 GuestQWML     [ lu  1   0] B07 Fla Sun Dec  6, 15:53 EST 2015
        history = []
        self.connection.history_owner = matchlist[0].groups()[0]
        for match in matchlist[2:]:
            # print(match.groups())
            history_no = match.groups()[0]
            result = match.groups()[1]
            owner_rating = match.groups()[2]
            owner_color = match.groups()[3]
            opp_rating = match.groups()[4]
            if result == "+":
                result = WHITEWON if owner_color == "W" else BLACKWON
            elif result == "-":
                result = WHITEWON if owner_color == "B" else BLACKWON
            else:
                result = DRAW
            opponent_name = match.groups()[5]
            if owner_color == "W":
                white = self.connection.history_owner
                black = opponent_name
                wrating = owner_rating
                brating = opp_rating
            else:
                white = opponent_name
                black = self.connection.history_owner
                brating = owner_rating
                wrating = opp_rating
            game_type = match.groups()[6]
            minutes, gain = match.groups()[7:9]
            reason = reasons_dict[match.groups()[10]]
            week, month, day, hour, minute, timezone, year = match.groups()[11:18]
            gametime = datetime.datetime(
                int(year), months.index(month) + 1, int(day), int(hour), int(minute)
            )
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            owner_color = owner_color == "B" and BLACK or WHITE
            minutes = int(minutes)
            gain = int(gain)

            wplayer = self.connection.players.get(white)
            bplayer = self.connection.players.get(black)
            game = FICSHistoryGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                minutes=minutes,
                inc=gain,
                private=private,
                wrating=wrating,
                brating=brating,
                time=gametime,
                reason=reason,
                history_no=history_no,
                result=result,
            )

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("historyGameAdded", game)
            history.append(game)

        self.emit("onHistoryList", history)

    __onHistoryResponseYES.BLKCMD = BLKCMD_HISTORY

    def __onJournalResponseYES(self, matchlist):
        # Journal for User:
        #     White         Rating  Black         Rating  Type         ECO End Result
        # %01: tentacle      2291    larsa         2050    [ lr  1   2] D35 Rep 1/2-1/2
        # %02: larsa         2045    tentacle      2296    [ lr  1   2] A46 Res 0-1
        journal = []
        self.connection.journal_owner = matchlist[0].groups()[0]
        for match in matchlist[2:]:
            # print(match.groups())
            journal_no = match.groups()[0]
            result = match.groups()[10]
            result = reprResult.index(result)
            white = match.groups()[1]
            wrating = match.groups()[2]
            black = match.groups()[3]
            brating = match.groups()[4]
            game_type = match.groups()[5]
            minutes, gain = match.groups()[6:8]
            reason = reasons_dict[match.groups()[9]]
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            minutes = int(minutes)
            gain = int(gain)

            wplayer = self.connection.players.get(white)
            bplayer = self.connection.players.get(black)
            game = FICSJournalGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                minutes=minutes,
                inc=gain,
                private=private,
                wrating=wrating,
                brating=brating,
                reason=reason,
                journal_no=journal_no,
                result=result,
            )

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("journalGameAdded", game)
            journal.append(game)

        self.emit("onJournalList", journal)

    __onJournalResponseYES.BLKCMD = BLKCMD_JOURNAL

    def __onStoredResponseNO(self, match):
        self.connection.stored_owner = match.groups()[0]
        self.emit("onAdjournmentsList", [])

    __onStoredResponseNO.BLKCMD = BLKCMD_STORED

    def __onHistoryResponseNO(self, match):
        self.connection.history_owner = match.groups()[0]
        self.emit("onHistoryList", [])

    __onHistoryResponseNO.BLKCMD = BLKCMD_HISTORY

    def __onJournalResponseNO(self, match):
        self.connection.journal_owner = match.groups()[0]
        self.emit("onJournalList", [])

    __onJournalResponseNO.BLKCMD = BLKCMD_JOURNAL

    def __onAdjournedGameResigned(self, match):
        self.queryAdjournments()

    def __onCurGameEnded(self, bm, game):
        if game.result == ADJOURNED:
            self.queryAdjournments()
        elif game.result in (DRAW, WHITEWON, BLACKWON):
            self.queryHistory()

    def queryAdjournments(self, owner=None):
        if owner is None:
            self.connection.client.run_command("stored")
        else:
            self.connection.client.run_command("stored %s" % owner)

    def queryHistory(self, owner=None):
        if owner is None:
            self.connection.client.run_command("history")
        else:
            self.connection.client.run_command("history %s" % owner)

    def queryJournal(self, owner=None):
        if owner is None:
            self.connection.client.run_command("journal")
        else:
            self.connection.client.run_command("journal %s" % owner)

    def queryMoves(self, game):
        if isinstance(game, FICSHistoryGame):
            self.connection.client.run_command(
                f"smoves {self.connection.history_owner} {game.history_no}"
            )
        elif isinstance(game, FICSJournalGame):
            self.connection.client.run_command(
                f"smoves {self.connection.journal_owner} %{game.journal_no}"
            )
        else:
            self.connection.client.run_command(
                f"smoves {self.connection.stored_owner} {game.opponent.name}"
            )

    def examine(self, game):
        game.board = None
        self.connection.archived_examine = game
        if isinstance(game, FICSAdjournedGame):
            self.connection.client.run_command(
                f"examine {self.connection.stored_owner} {game.opponent.name}"
            )
        elif isinstance(game, FICSHistoryGame):
            self.connection.client.run_command(
                f"examine {self.connection.history_owner} {game.history_no}"
            )
        elif isinstance(game, FICSJournalGame):
            self.connection.client.run_command(
                f"examine {self.connection.journal_owner} %{game.journal_no}"
            )

    def challenge(self, playerName):
        self.connection.client.run_command("match %s" % playerName)

    def resign(self, game):
        """This is (and draw and abort) are possible even when one's
        opponent is not logged on"""
        if not game.opponent.adjournment:
            log.warning(
                "AdjournManager.resign: no adjourned game vs %s" % game.opponent
            )
            return
        log.info("AdjournManager.resign: resigning adjourned game=%s" % game)
        self.connection.client.run_command("resign %s" % game.opponent.name)

    def draw(self, game):
        if not game.opponent.adjournment:
            log.warning("AdjournManager.draw: no adjourned game vs %s" % game.opponent)
            return
        log.info("AdjournManager.draw: offering sdraw for adjourned game=%s" % game)
        self.connection.client.run_command("sdraw %s" % game.opponent.name)

    def abort(self, game):
        if not game.opponent.adjournment:
            log.warning("AdjournManager.abort: no adjourned game vs %s" % game.opponent)
            return
        log.info("AdjournManager.abort: offering sabort for adjourned game=%s" % game)
        self.connection.client.run_command("sabort %s" % game.opponent.name)

    def resume(self, game):
        if not game.opponent.adjournment:
            log.warning(
                "AdjournManager.resume: no adjourned game vs %s" % game.opponent
            )
            return
        log.info("AdjournManager.resume: offering resume for adjourned game=%s" % game)
        self.connection.client.run_command("match %s" % game.opponent.name)

    # (a)  Users who have more than 15 stored games are restricted from starting new
    # games.  If this situation happens to you, review your stored games and see
    # which ones might be eligible for adjudication (see "help adjudication").



from math import pi, cos, sin

import cairo
from gi.repository import GLib, Gtk, Gdk, Pango, PangoCairo, GObject

from pychess.System import conf
from pychess.Utils import formatTime
from pychess.Utils.const import (
    BLACK,
    WHITE,
    LOCAL,
    UNFINISHED_STATES,
    DRAW,
    WHITEWON,
    BLACKWON,
    UNKNOWN_STATE,
)
from . import preferencesDialog


class ChessClock(Gtk.DrawingArea):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.connect("draw", self.expose)
        self.names = [_("White"), _("Black")]

        self.model = None
        self.short_on_time = [False, False]
        self.alarm_spin = conf.get("alarm_spin")

        conf.notify_add("alarm_spin", self.on_alarm_spin)

    def on_alarm_spin(self, *args):
        self.alarm_spin = conf.get("alarm_spin")

    def expose(self, widget, ctx):
        context = widget.get_window().cairo_create()
        clip_ext = context.clip_extents()

        if clip_ext[0] > 0 or clip_ext[2] < self.get_allocated_width():
            self.redraw_canvas()
            return False

        rec = Gdk.Rectangle()
        rec.x, rec.y, rec.width, rec.height = (
            clip_ext[0],
            clip_ext[1],
            clip_ext[2] - clip_ext[0],
            clip_ext[3] - clip_ext[1],
        )
        context.rectangle(rec.x, rec.y, rec.width, rec.height)
        context.clip()
        self.draw(context)
        return False

    def draw(self, context):
        style_ctxt = self.get_style_context()
        self.light = style_ctxt.lookup_color("p_light_color")[1]
        self.dark = style_ctxt.lookup_color("p_dark_color")[1]
        if not self.model:
            return

        # Draw graphical Clock. Should this be moved to preferences?
        drawClock = True

        rect = Gdk.Rectangle()
        clip_ext = context.clip_extents()
        rect.x, rect.y, rect.width, rect.height = (
            clip_ext[0],
            clip_ext[1],
            clip_ext[2] - clip_ext[0],
            clip_ext[3] - clip_ext[1],
        )
        context.rectangle(
            rect.width / 2.0 * self.model.movingColor, 0, rect.width / 2.0, rect.height
        )

        context.set_source_rgba(
            self.dark.red, self.dark.green, self.dark.blue, self.dark.alpha
        )
        context.fill_preserve()
        context.new_path()

        time0 = self.names[0], self.formatedCache[WHITE]
        layout0 = self.create_pango_layout(" %s: %s " % (time0))
        layout0.set_font_description(Pango.FontDescription("Sans Serif 17"))

        time1 = self.names[1], self.formatedCache[BLACK]
        layout1 = self.create_pango_layout(" %s: %s " % (time1))
        layout1.set_font_description(Pango.FontDescription("Sans Serif 17"))

        dbl_max = max(layout1.get_pixel_size()[0], layout0.get_pixel_size()[0]) * 2
        self.set_size_request(dbl_max + rect.height + 7, -1)

        pangoScale = float(Pango.SCALE)

        # Analog clock code.
        def paintClock(player):
            clock_y = rect.height / 2.0
            clock_x = clock_y + rect.width / 2.0 * player + 1
            rec = rect.height / 2.0 - 3.5

            context.arc(clock_x, clock_y, rec - 1, 0, 2 * pi)
            linear = cairo.LinearGradient(
                clock_x - rec * 2,
                clock_y - rec * 2,
                clock_x + rec * 2,
                clock_y + rec * 2,
            )
            linear.add_color_stop_rgba(0, 1, 1, 1, 0.3)
            linear.add_color_stop_rgba(1, 0, 0, 0, 0.3)
            # context.set_source_rgba( 0, 0, 0, .3)
            context.set_source(linear)
            context.fill()

            linear = cairo.LinearGradient(
                clock_x - rec, clock_y - rec, clock_x + rec, clock_y + rec
            )
            linear.add_color_stop_rgba(0, 0, 0, 0, 0.5)
            linear.add_color_stop_rgba(1, 1, 1, 1, 0.5)
            context.arc(clock_x, clock_y, rec, 0, 2 * pi)
            context.set_source(linear)
            context.set_line_width(2.5)
            context.stroke()

            starttime = float(self.model.getInitialTime()) or 1
            used = self.model.getPlayerTime(player) / starttime
            if used > 0:
                if used > 0:
                    context.arc(
                        clock_x, clock_y, rec - 0.8, -(used + 0.25) * 2 * pi, -0.5 * pi
                    )
                    context.line_to(clock_x, clock_y)
                    context.close_path()
                elif used == 0:
                    context.arc(clock_x, clock_y, rec - 0.8, -0.5 * pi, 1.5 * pi)
                    context.line_to(clock_x, clock_y)

                radial = cairo.RadialGradient(
                    clock_x, clock_y, 3, clock_x, clock_y, rec
                )
                if player == 0:
                    # radial.add_color_stop_rgb(0, .73, .74, .71)
                    radial.add_color_stop_rgb(0, 0.93, 0.93, 0.92)
                    radial.add_color_stop_rgb(1, 1, 1, 1)
                else:
                    # radial.add_color_stop_rgb(0, .53, .54, .52)
                    radial.add_color_stop_rgb(0, 0.18, 0.20, 0.21)
                    radial.add_color_stop_rgb(1, 0, 0, 0)
                context.set_source(radial)
                context.fill()

                x_loc = clock_x - cos((used - 0.25) * 2 * pi) * (rec - 1)
                y_loc = clock_y + sin((used - 0.25) * 2 * pi) * (rec - 1)
                context.move_to(clock_x, clock_y - rec + 1)
                context.line_to(clock_x, clock_y)
                context.line_to(x_loc, y_loc)
                context.set_line_width(0.2)
                if player == 0:
                    context.set_source_rgb(0, 0, 0)
                else:
                    context.set_source_rgb(1, 1, 1)
                context.stroke()

        if drawClock:
            paintClock(WHITE)
        if (self.model.movingColor or WHITE) == WHITE:
            context.set_source_rgba(
                self.light.red, self.light.green, self.light.blue, self.light.alpha
            )
        else:
            context.set_source_rgba(
                self.dark.red, self.dark.green, self.dark.blue, self.dark.alpha
            )
        y_loc = (
            rect.height / 2.0
            - layout0.get_extents()[0].height / pangoScale / 2
            - layout0.get_extents()[0].y / pangoScale
        )
        context.move_to(rect.height - 7, y_loc)
        PangoCairo.show_layout(context, layout0)

        if drawClock:
            paintClock(BLACK)
        if self.model.movingColor == BLACK:
            context.set_source_rgba(
                self.light.red, self.light.green, self.light.blue, self.light.alpha
            )
        else:
            context.set_source_rgba(
                self.dark.red, self.dark.green, self.dark.blue, self.dark.alpha
            )
        y_loc = (
            rect.height / 2.0
            - layout0.get_extents()[0].height / pangoScale / 2
            - layout0.get_extents()[0].y / pangoScale
        )
        context.move_to(rect.width / 2.0 + rect.height - 7, y_loc)
        PangoCairo.show_layout(context, layout1)

    def redraw_canvas(self):
        def do_redraw_canvas():
            if self.get_window():
                allocation = self.get_allocation()
                rect = Gdk.Rectangle()
                rect.x, rect.y, rect.width, rect.height = (
                    0,
                    0,
                    allocation.width,
                    allocation.height,
                )
                self.get_window().invalidate_rect(rect, True)
                self.get_window().process_updates(True)

        GLib.idle_add(do_redraw_canvas)

    def setModel(self, model):
        self.model = model
        self.model.connect("time_changed", self.time_changed)
        self.model.connect("player_changed", self.player_changed)
        self.formatedCache = [
            formatTime(self.model.getPlayerTime(self.model.movingColor or WHITE))
        ] * 2
        if model.secs != 0 or model.gain != 0:
            GLib.timeout_add(100, self.update)

    def time_changed(self, model):
        self.update()

    def player_changed(self, model):
        self.redraw_canvas()

    def update(self, wmovecount=-1, bmovecount=-1):
        if self.model.ended:
            return False
        if len(self.model.gamemodel.players) < 2:
            return not self.model.ended
        alarm_time = int(self.alarm_spin)
        if (
            self.model.getPlayerTime(self.model.movingColor) <= alarm_time
            and self.model.gamemodel.players[self.model.movingColor].__type__ == LOCAL
            and self.model.gamemodel.status in UNFINISHED_STATES
            and self.model.secs != 0
            and self.model.gamemodel.endstatus
            not in (DRAW, WHITEWON, BLACKWON, UNKNOWN_STATE)
            and not self.short_on_time[self.model.movingColor]
        ):
            self.short_on_time[self.model.movingColor] = True
            preferencesDialog.SoundTab.playAction("shortOnTime")

        if self.model.paused and wmovecount == -1 and bmovecount == -1:
            return not self.model.ended
        white_time = formatTime(self.model.getPlayerTime(WHITE, wmovecount))
        black_time = formatTime(self.model.getPlayerTime(BLACK, bmovecount))
        if self.formatedCache != [white_time, black_time]:
            self.formatedCache = [white_time, black_time]
            self.redraw_canvas()
        return not self.model.ended



from gi.repository import GObject

sanmove = "([a-hx@OoPKQRBN0-8+#=-]{2,7})"


class ErrorManager(GObject.GObject):
    __gsignals__ = {
        "onCommandNotFound": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "onAmbiguousMove": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "onIllegalMove": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        connection.expect_line(self.onError, r"(.*?): Command not found\.")
        connection.expect_line(
            self.onAmbiguousMove, r"Ambiguous move \((%s)\)\." % sanmove
        )
        connection.expect_line(self.onIllegalMove, r"Illegal move \((%s)\)\." % sanmove)

    def onError(self, match):
        command = match.groups()[0]
        self.emit("onCommandNotFound", command)

    def onAmbiguousMove(self, match):
        move = match.groups()[0]
        self.emit("onAmbiguousMove", move)

    def onIllegalMove(self, match):
        move = match.groups()[0]
        self.emit("onIllegalMove", move)



import asyncio

from gi.repository import Gtk

from pychess.Utils import wait_signal
from pychess.System import uistuff


class DiscovererDialog:
    def __init__(self, discoverer):
        self.discoverer = discoverer
        self.widgets = uistuff.GladeWidgets("discovererDialog.glade")

        # =======================================================================
        # Clear glade defaults
        # =======================================================================
        for child in self.widgets["enginesTable"].get_children():
            self.widgets["enginesTable"].remove(child)

        self.finished = False
        self.throbber = None
        self.nameToBar = {}

        discoverer.pre_discover()
        binnames = self.discoverer.toBeRechecked.keys()
        if len(binnames) == 0:
            self.finished = True
        # ======================================================================
        # Insert the names to be discovered
        # ======================================================================
        for i, name in enumerate(binnames):
            label = Gtk.Label(label=name + ":")
            label.props.xalign = 1
            self.widgets["enginesTable"].attach(label, 0, 1, i, i + 1)
            bar = Gtk.ProgressBar()
            self.widgets["enginesTable"].attach(bar, 1, 2, i, i + 1)
            self.nameToBar[name] = bar

        # =======================================================================
        # Add throbber
        # =======================================================================

        self.throbber = Gtk.Spinner()
        self.throbber.set_size_request(50, 50)
        self.widgets["throbberDock"].add(self.throbber)

        # =======================================================================
        # Show the window
        # =======================================================================
        self.widgets["discovererDialog"].set_position(
            Gtk.WindowPosition.CENTER_ON_PARENT
        )
        self.widgets["discovererDialog"].show_all()
        self.throbber.start()

    async def start(self):
        if self.finished:
            self.close()

        # let dialog window draw itself
        await asyncio.sleep(0.1)

        asyncio.create_task(self.all_whatcher())
        asyncio.create_task(self.discovered_whatcher())

        self.discoverer.discover()

    async def discovered_whatcher(self):
        while True:
            if self.finished:
                return

            _discoverer, binname, _xmlenginevalue = await wait_signal(
                self.discoverer, "engine_discovered"
            )

            if binname in self.nameToBar:
                bar = self.nameToBar[binname]
                bar.props.fraction = 1

    async def all_whatcher(self):
        await wait_signal(self.discoverer, "all_engines_discovered")

        self.finished = True
        self.close()

    def close(self):
        if self.throbber:
            self.throbber.stop()
        self.widgets["discovererDialog"].hide()



from gi.repository import Gtk


def getUserTextDialog(parent, title, description):
    dialog = Gtk.Dialog(
        title,
        parent,
        Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK,
            Gtk.ResponseType.ACCEPT,
        ),
    )
    dialog.set_resizable(False)
    dialog.set_size_request(400, -1)

    vbx = Gtk.VBox()
    label = Gtk.Label(description)
    label.set_xalign(0)
    vbx.pack_start(label, True, True, 5)
    textedit = Gtk.Entry()
    textedit.connect("activate", lambda p: dialog.response(Gtk.ResponseType.ACCEPT))
    vbx.pack_end(textedit, True, True, 5)

    hbx = Gtk.HBox()
    hbx.pack_start(vbx, True, True, 10)

    dialog.get_content_area().pack_start(hbx, True, True, 5)
    dialog.show_all()
    if dialog.run() == Gtk.ResponseType.ACCEPT:
        result = textedit.get_text()
    else:
        result = None
    dialog.destroy()
    return result



from gi.repository import GObject


class PlayerIsDead(Exception):
    """Used instead of returning a move,
    when an engine crashes, or a nonlocal player disconnects"""

    pass


class PassInterrupt(Exception):
    """Used instead of returning a move, when a players turn is interrupted but not changed.
    This may happen when undoMoves doesn't changes the current player"""

    pass


class TurnInterrupt(Exception):
    """Used instead of returning a move, when a players turn is interrupted.
    This may happen when undoMoves changes the current player"""

    pass


class InvalidMove(Exception):
    """Used instead of returning a move,
    when an engine plays an invalid move"""

    pass


class GameEnded(Exception):
    """Used instead of returning a move on game end"""

    pass


class Player(GObject.GObject):
    __gsignals__ = {
        "offer": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "withdraw": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "decline": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "accept": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "name_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.name = ""
        self.ichandle = None
        self.icrating = None

    def setName(self, name):
        """__repr__ should return this name"""
        self.name = name
        self.emit("name_changed")

    def __repr__(self):
        return self.name

    @property
    def time(self):
        pass  # Optional

    # Starting the game

    def prestart(self):
        pass  # Optional

    def start(self, event, is_dead):
        event.set()

    def setOptionInitialBoard(self, model):
        pass  # Optional. Further defined in Engine.py

    # Ending the game

    def end(self, status, reason):
        """Called when the game ends in a normal way. Use this for shutting
        down engines etc."""
        raise NotImplementedError

    def kill(self, reason):
        """Called when game has too die fast and ugly. Mostly used in case of
        errors and stuff. Use for closing connections etc."""
        raise NotImplementedError

    # Send the player move updates

    def makeMove(self, board1, move, board2):
        """Takes a board object, and if ply>lowply the latest move object and
        second latest board object as well. Otherwise these two are None.
        Retruns: A new move object, witch the player wants to do."""
        raise NotImplementedError

    def putMove(self, board1, move, board2):
        """Like makeMove, but doesn't block and doesn't return anything.
        putMove is only used when the player is spectatctor to a game"""
        # Optional

    def updateTime(self, secs, opsecs):
        """Updates the player with the current remaining time as a float of
        seconds"""
        # Optional

    # nteracting with the player

    def pause(self):
        """Should stop the player from thinking until resume is called"""
        raise NotImplementedError

    def resume(self):
        """Should resume player to think if he's paused"""
        raise NotImplementedError

    def hurry(self):
        """Forces engines to move now, and sends a hurry message to nonlocal
        human players"""
        # Optional

    def undoMoves(self, moves, gamemodel):
        """Undo 'moves' moves and makes the latest board in gamemodel the
        current"""
        # Optional

    def playerUndoMoves(self, moves, gamemodel):
        """Some players undo different depending on whether they are players or
        spectators. This is a convenient way to handle that."""
        # Optional
        return self.undoMoves(moves, gamemodel)

    def spectatorUndoMoves(self, moves, gamemodel):
        """Some players undo different depending on whether they are players or
        spectators. This is a convenient way to handle that."""
        # Optional
        return self.undoMoves(moves, gamemodel)

    def putMessage(self, message):
        """Sends the player a chatmessage"""
        # Optional

    # Offer handling

    def offer(self, offer):
        """The players opponent has offered the player offer. If the player
        accepts, it should respond by mirroring the offer with
        emit("accept", offer). If it should either ignore the offer or emit
        "decline"."""
        raise NotImplementedError

    def offerDeclined(self, offer):
        """An offer sent by the player was responded negative by the
        opponent"""
        # Optional

    def offerWithdrawn(self, offer):
        """An offer earlier offered to the player has been withdrawn"""
        # Optional

    def offerError(self, offer, error):
        """An offer, accept or action made by the player has been refused by
        the game model."""
        # Optional



import asyncio
import itertools
import re

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject

from pychess.Utils import wait_signal
from pychess.System import conf
from pychess.System.Log import log
from pychess.Utils.Move import Move
from pychess.Utils.Board import Board
from pychess.Utils.Cord import Cord
from pychess.Utils.Move import toSAN, toAN, parseAny
from pychess.Utils.Offer import Offer
from pychess.Utils.const import (
    ANALYZING,
    INVERSE_ANALYZING,
    DRAW,
    WHITEWON,
    BLACKWON,
    WON_ADJUDICATION,
    DRAW_OFFER,
    ACTION_ERROR_NONE_TO_ACCEPT,
    CASTLE_KK,
    WHITE,
    CASTLE_SAN,
    FISCHERRANDOMCHESS,
    BLACK,
    reprSign,
    RESIGNATION,
    SCHESS,
)
from pychess.Utils.logic import validate, getMoveKillingKing
from pychess.Utils.lutils.ldata import MATE_VALUE
from pychess.Utils.lutils.lmove import ParsingError
from pychess.Variants import variants
from pychess.Players.Player import PlayerIsDead, TurnInterrupt, InvalidMove
from .ProtocolEngine import ProtocolEngine, TIME_OUT_SECOND


movere = re.compile(
    r"""
    (                   # group start
    (?:                 # non grouping parenthesis start
    [PKQRBN]?            # piece
    [a-h]?[1-8]?        # unambiguous column or line
    x?                  # capture
    @?                  # drop
    [a-h][1-8]          # destination square
    =?[QRBN]?           # promotion
    |O\-O(?:\-O)?       # castling
    |0\-0(?:\-0)?       # castling
    )                   # non grouping parenthesis end
    [+#]?               # check/mate
    )                   # group end
    \s*                 # any whitespace
    """,
    re.VERBOSE,
)

d_plus_dot_expr = re.compile(r"\d+\.")

anare = re.compile(
    r"""
    ^                        # beginning of string
    (\s*                     #
    \d+ [+\-\.]?             # The ply analyzed. Some engines end it with a dot, minus or plus
    \s+)                     #
    (-?Mat\s*\d+ | [+\-\d\.]+) # The score found in centipawns.
                             #   Mat1 is used by gnuchess to specify mate in one.
                             #   otherwise we should support a signed float
    \s+                      #
    ([\d\.]+)                # The time used in centi-seconds
    \s+                      #
    ([\d\.]+)                # Number of nodes visited
    \s+                      #
    (.+)                     # The Principal-Variation. With or without move numbers
    \s*                      #
    $                        # end of string
    """,
    re.VERBOSE,
)

# anare = re.compile("\(d+)\.?\s+ (Mat\d+|[-\d\.]+) \s+ \d+\s+\d+\s+((?:%s\s*)+)" % mov)

whitespaces = re.compile(r"\s+")


# There is no way in the CECP protocol to determine if an engine not answering
# the protover=2 handshake with done=1 is old or just very slow. Thus we
# need a timeout after which we conclude the engine is 'protover=1' and will
# never answer.
# XBoard will only give 2 seconds, but as we are quite sure that
# the engines support the protocol, we can add more. We don't add
# infinite time though, just in case.
# The engine can get more time by sending done=0


class CECPEngine(ProtocolEngine):
    def __init__(self, subprocess, color, protover, md5):
        ProtocolEngine.__init__(self, subprocess, color, protover, md5)

        self.features = {
            "ping": 0,
            "setboard": 0,
            "playother": 0,
            "san": 0,
            "usermove": 0,
            "time": 1,
            "draw": 1,
            "sigint": 0,
            "sigterm": 0,
            "reuse": 0,
            "analyze": 0,
            "myname": ", ".join(self.defname),
            "variants": None,
            "colors": 1,
            "ics": 0,
            "name": 0,
            "pause": 0,
            "nps": 0,
            "debug": 0,
            "memory": 0,
            "smp": 0,
            "egt": "",
            "option": "",
            "exclude": 0,
            "done": None,
        }

        self.supported_features = [
            "ping",
            "setboard",
            "san",
            "usermove",
            "time",
            "draw",
            "sigint",
            "analyze",
            "myname",
            "variants",
            "colors",
            "pause",
            "done",
            "egt",
            "debug",
            "smp",
            "memory",
            "option",
        ]

        self.options = {}
        self.options["Ponder"] = {"name": "Ponder", "type": "check", "default": False}

        self.name = None

        self.board = Board(setup=True)

        # if self.engineIsInNotPlaying == True, engine is in "force" mode,
        # i.e. not thinking or playing, but still verifying move legality
        self.engineIsInNotPlaying = False
        self.engineIsAnalyzing = False
        self.movenext = False
        self.waitingForMove = False
        self.readyForMoveNowCommand = False
        self.timeHandicap = 1

        self.lastping = 0
        self.lastpong = 0

        self.queue = asyncio.Queue()
        self.parse_line_task = asyncio.create_task(self.parseLine(self.engine))
        self.died_cid = self.engine.connect(
            "died", lambda e: self.queue.put_nowait("die")
        )
        self.invalid_move = None

        self.optionQueue = []
        self.undoQueue = []
        self.ready_moves_event = asyncio.Event()

        self.cids = [
            self.connect_after("readyForOptions", self.__onReadyForOptions),
            self.connect_after("readyForMoves", self.__onReadyForMoves),
        ]

    # Starting the game

    def prestart(self):
        print("xboard", file=self.engine)
        if self.protover == 1:
            # start a new game (CECPv1 engines):
            print("new", file=self.engine)

            # we are now ready for options:
            self.emit("readyForOptions")
        elif self.protover == 2:
            # start advanced protocol initialisation:
            print("protover 2", file=self.engine)

            # we don't start a new game for CECPv2 here,
            # we will do it after feature accept/reject is completed.

    def start(self, event, is_dead):
        asyncio.create_task(self.__startBlocking(event, is_dead))

    async def __startBlocking(self, event, is_dead):
        if self.protover == 1:
            self.emit("readyForMoves")
            return_value = "ready"

        if self.protover == 2:
            try:
                return_value = await asyncio.wait_for(self.queue.get(), TIME_OUT_SECOND)
                if return_value == "not ready":
                    return_value = await asyncio.wait_for(
                        self.queue.get(), TIME_OUT_SECOND
                    )
                    # Gaviota sends done=0 after "xboard" and after "protover 2" too
                    if return_value == "not ready":
                        return_value = await asyncio.wait_for(
                            self.queue.get(), TIME_OUT_SECOND
                        )
            except asyncio.TimeoutError:
                log.warning("Got timeout error", extra={"task": self.defname})
                is_dead.add(True)
            except Exception:
                log.warning("Unknown error", extra={"task": self.defname})
                is_dead.add(True)
            else:
                if return_value == "die":
                    is_dead.add(True)
                assert return_value == "ready" or return_value == "del"

        if event is not None:
            event.set()

    def __onReadyForOptions(self, self_):
        # We always want post turned on so the Engine Output sidebar can
        # show those things  -Jonas Thiem
        print("post", file=self.engine)

        for command in self.optionQueue:
            print(command, file=self.engine)

    def __onReadyForMoves(self, self_):
        if self.mode in (ANALYZING, INVERSE_ANALYZING):
            # workaround for crafty not sending analysis after it has found a mating line
            # http://code.google.com/p/pychess/issues/detail?id=515
            if "crafty" in self.features["myname"].lower():
                print("noise 0", file=self.engine)

            self.__sendAnalyze(self.mode == INVERSE_ANALYZING)
        self.ready_moves_event.set()
        self.readyMoves = True

    # Ending the game

    def end(self, status, reason):
        self.parse_line_task.cancel()
        if self.engine.handler_is_connected(self.died_cid):
            self.engine.disconnect(self.died_cid)
        if self.handler_is_connected(self.analyze_cid):
            self.disconnect(self.analyze_cid)
        for cid in self.cids:
            if self.handler_is_connected(cid):
                self.disconnect(cid)
        self.board = None

        if self.connected:
            # We currently can't fillout the comment "field" as the repr strings
            # for reasons and statuses lies in Main.py
            # Creating Status and Reason class would solve this
            if status == DRAW:
                print("result 1/2-1/2 {?}", file=self.engine)
            elif status == WHITEWON:
                print("result 1-0 {?}", file=self.engine)
            elif status == BLACKWON:
                print("result 0-1 {?}", file=self.engine)
            else:
                print("result * {?}", file=self.engine)

            if reason == WON_ADJUDICATION:
                self.queue.put_nowait("invalid")

                # Make sure the engine exits and do some cleaning
            self.kill(reason)

    def kill(self, reason):
        """Kills the engine, starting with the 'quit' command, then sigterm and
        eventually sigkill.
        Returns the exitcode, or if engine have already been killed, returns
        None"""
        if self.connected:
            self.connected = False
            try:
                try:
                    print("quit", file=self.engine)
                    self.queue.put_nowait("del")
                    self.engine.terminate()

                except OSError as err:
                    # No need to raise on a hang up error, as the engine is dead
                    # anyways
                    if err.errno == 32:
                        log.warning("Hung up Error", extra={"task": self.defname})
                        return err.errno
                    else:
                        raise

            finally:
                # Clear the analyzed data, if any
                self.emit("analyze", [])

    # Send the player move updates

    def set_board(self, board):
        self.setBoardList([board], [])

    def setBoard(self, board, search=True):
        async def coro():
            if self.engineIsAnalyzing:
                self.__stop_analyze()
                await asyncio.sleep(0.1)

            self.setBoardList([board], [])
            if search:
                self.__sendAnalyze(self.mode == INVERSE_ANALYZING)

        asyncio.create_task(coro())

    def putMove(self, board1, move, board2):
        """Sends the engine the last move made (for spectator engines).
        @param board1: The current board
        @param move: The last move made
        @param board2: The board before the last move was made
        """

        async def coro():
            if self.engineIsAnalyzing:
                self.__stop_analyze()
                await asyncio.sleep(0.1)

            self.setBoardList([board1], [])
            if not self.analyzing_paused:
                self.__sendAnalyze(self.mode == INVERSE_ANALYZING)

        asyncio.create_task(coro())

    async def makeMove(self, board1, move, board2):
        """Gets a move from the engine (for player engines).
        @param board1: The current board
        @param move: The last move made
        @param board2: The board before the last move was made
        @return: The move the engine decided to make
        """
        log.debug(
            "makeMove: move={} self.movenext={} board1={} board2={} self.board={}".format(
                move, self.movenext, board1, board2, self.board
            ),
            extra={"task": self.defname},
        )
        assert self.readyMoves

        if self.board == board1 or not board2 or self.movenext:
            self.board = board1
            self.__tellEngineToPlayCurrentColorAndMakeMove()
            self.movenext = False
        else:
            self.board = board1
            self.__usermove(board2, move)

            if self.engineIsInNotPlaying:
                self.__tellEngineToPlayCurrentColorAndMakeMove()

        self.waitingForMove = True
        self.readyForMoveNowCommand = True

        # Parse outputs
        status = await self.queue.get()
        if status == "not ready":
            log.warning(
                "Engine seems to be protover=2, but is treated as protover=1",
                extra={"task": self.defname},
            )
            status = await self.queue.get()
        if status == "ready":
            status = await self.queue.get()
        if status == "invalid":
            raise InvalidMove
        if status == "del" or status == "die":
            raise PlayerIsDead("Killed by foreign forces")
        if status == "int":
            raise TurnInterrupt

        self.waitingForMove = False
        self.readyForMoveNowCommand = False
        assert isinstance(status, Move), status
        return status

    def updateTime(self, secs, opsecs):
        if self.features["time"]:
            print("time %s" % int(secs * 100 * self.timeHandicap), file=self.engine)
            print("otim %s" % int(opsecs * 100), file=self.engine)

    # Standard options

    def setOptionAnalyzing(self, mode):
        self.mode = mode

    def setOptionInitialBoard(self, model):
        async def coro():
            await self.ready_moves_event.wait()
            # We don't use the optionQueue here, as set board prints a whole lot of
            # stuff. Instead we just call it.
            self.setBoardList(model.boards[:], model.moves[:])

        asyncio.create_task(coro())

    def setBoardList(self, boards, moves):
        # Notice: If this method is to be called while playing, the engine will
        # need 'new' and an arrangement similar to that of 'pause' to avoid
        # the current thought move to appear
        if self.mode not in (ANALYZING, INVERSE_ANALYZING):
            self.__tellEngineToStopPlayingCurrentColor()

        self.__setBoard(boards[0])

        self.board = boards[-1]
        for board, move in zip(boards[:-1], moves):
            self.__usermove(board, move)

        if self.mode in (ANALYZING, INVERSE_ANALYZING):
            self.board = boards[-1]
        if self.mode == INVERSE_ANALYZING:
            self.board = self.board.switchColor()

        # The called of setBoardList will have to repost/analyze the
        # analyzer engines at this point.

    def setOptionVariant(self, variant):
        if self.features["variants"] is None:
            log.warning(
                "setOptionVariant: engine doesn't support variants",
                extra={"task": self.defname},
            )
            return

        if variant in variants.values() and not variant.standard_rules:
            assert (
                variant.cecp_name in self.features["variants"]
            ), f"{self} doesn't support {variant.cecp_name} variant"
            self.optionQueue.append("variant %s" % variant.cecp_name)

    #    Strength system                               #
    #          Strength  Depth  Ponder  Time handicap  #
    #          1         1      o       1,258%         #
    #          2         2      o       1,584%         #
    #          3         3      o       1.995%         #
    #                                                  #
    #         19         o      x       79,43%         #
    #         20         o      x       o              #

    def setOptionStrength(self, strength, forcePonderOff):
        self.strength = strength

        if strength <= 19:
            self.__setTimeHandicap(0.01 * 10 ** (strength / 10.0))

        if strength <= 18:
            self.__setDepth(strength)

        # Crafty ofers 100 skill levels
        if "crafty" in self.features["myname"].lower() and strength <= 19:
            self.optionQueue.append("skill %s" % strength * 5)

        self.__setPonder(strength >= 19 and not forcePonderOff)

        if strength == 20:
            if "gaviota" in self.features["egt"]:
                self.optionQueue.append("egtpath gaviota %s" % conf.get("egtb_path"))
        else:
            self.optionQueue.append("random")

    def __setDepth(self, depth):
        self.optionQueue.append("sd %d" % depth)

    def __setTimeHandicap(self, timeHandicap):
        self.timeHandicap = timeHandicap

    def __setPonder(self, ponder):
        if ponder:
            self.optionQueue.append("hard")
        else:
            self.optionQueue.append("hard")
            self.optionQueue.append("easy")

    def setOptionTime(self, secs, gain, moves):
        # Notice: In CECP we apply time handicap in updateTime, not in
        #         setOptionTime.

        minutes = int(secs / 60)
        secs = int(secs % 60)
        mins = str(minutes)
        if secs:
            mins += ":" + str(secs)

        self.optionQueue.append("level %s %s %d" % (moves, mins, gain))

    # Option handling

    def setOption(self, key, value):
        """Set an option, which will be sent to the engine, after the
        'readyForOptions' signal has passed.
        If you want to know the possible options, you should go to
        engineDiscoverer or use the hasOption method
        while you are in your 'readyForOptions' signal handler"""
        if self.readyMoves:
            log.warning(
                "Options set after 'readyok' are not sent to the engine",
                extra={"task": self.defname},
            )
        if key == "cores":
            self.optionQueue.append("cores %s" % value)
        elif key == "memory":
            self.optionQueue.append("memory %s" % value)
        elif key.lower() == "ponder":
            self.__setPonder(value == 1)
        else:
            self.optionQueue.append(f"option {key}={value}")

    # Interacting with the player

    def pause(self):
        """Pauses engine using the "pause" command if available. Otherwise put
        engine in force mode. By the specs the engine shouldn't ponder in
        force mode, but some of them do so anyways."""

        log.debug("pause: self=%s" % self, extra={"task": self.defname})
        if self.isAnalyzing():
            self.__stop_analyze()
            self.analyzing_paused = True
        else:
            self.engine.pause()
        return

    def resume(self):
        log.debug("resume: self=%s" % self, extra={"task": self.defname})
        if self.isAnalyzing():
            self.__sendAnalyze(self.mode == INVERSE_ANALYZING)
            self.analyzing_paused = False
        else:
            self.engine.resume()
        return

    def hurry(self):
        log.debug(
            "hurry: self.waitingForMove={} self.readyForMoveNowCommand={}".format(
                self.waitingForMove, self.readyForMoveNowCommand
            ),
            extra={"task": self.defname},
        )
        if self.waitingForMove and self.readyForMoveNowCommand:
            self.__tellEngineToMoveNow()
            self.readyForMoveNowCommand = False

    def spectatorUndoMoves(self, moves, gamemodel):
        if self.analyzing_paused:
            return

        log.debug(
            "spectatorUndoMoves: moves={} gamemodel.ply={} gamemodel.boards[-1]={} self.board={}".format(
                moves, gamemodel.ply, gamemodel.boards[-1], self.board
            ),
            extra={"task": self.defname},
        )

        for i in range(moves):
            print("undo", file=self.engine)

        self.board = gamemodel.boards[-1]

    def playerUndoMoves(self, moves, gamemodel):
        log.debug(
            "CECPEngine.playerUndoMoves: moves=%s self=%s gamemodel.curplayer=%s"
            % (moves, self, gamemodel.curplayer),
            extra={"task": self.defname},
        )

        self.board = gamemodel.boards[-1]

        self.__tellEngineToStopPlayingCurrentColor()

        for i in range(moves):
            print("undo", file=self.engine)

        if (
            gamemodel.curplayer != self
            and moves % 2 == 1
            or (gamemodel.curplayer == self and moves % 2 == 0)
        ):
            # Interrupt if we were searching, but should no longer do so
            log.debug(
                "CECPEngine.playerUndoMoves: putting TurnInterrupt into self.move_queue %s"
                % self.name,
                extra={"task": self.defname},
            )
            self.queue.put_nowait("int")

    # Offer handling

    def offer(self, offer):
        if offer.type == DRAW_OFFER:
            if self.features["draw"]:
                print("draw", file=self.engine)
        else:
            self.emit("accept", offer)

    def offerError(self, offer, error):
        if self.features["draw"]:
            # We don't keep track if engine draws are offers or accepts. We just
            # Always assume they are accepts, and if they are not, we get this
            # error and emit offer instead
            if offer.type == DRAW_OFFER and error == ACTION_ERROR_NONE_TO_ACCEPT:
                self.emit("offer", Offer(DRAW_OFFER))

    # Internal

    def __usermove(self, board, move):
        if self.features["usermove"]:
            self.engine.write("usermove ")

        if self.features["san"]:
            print(toSAN(board, move), file=self.engine)
        else:
            castle_notation = CASTLE_KK
            if board.variant == FISCHERRANDOMCHESS:
                castle_notation = CASTLE_SAN
            print(
                toAN(board, move, short=True, castleNotation=castle_notation),
                file=self.engine,
            )

    def __tellEngineToMoveNow(self):
        if self.features["sigint"]:
            self.engine.sigint()
        print("?", file=self.engine)

    def __tellEngineToStopPlayingCurrentColor(self):
        print("force", file=self.engine)
        self.engineIsInNotPlaying = True

    def __tellEngineToPlayCurrentColorAndMakeMove(self):
        self.__printColor()
        print("go", file=self.engine)
        self.engineIsInNotPlaying = False

    def __stop_analyze(self):
        if self.engineIsAnalyzing:
            print("exit", file=self.engine)
            # Some engines (crafty, gnuchess) doesn't respond to exit command
            # we try to force them to stop with an empty board fen
            print("setboard 8/8/8/8/8/8/8/8 w - - 0 1", file=self.engine)
            self.engineIsAnalyzing = False

    def __sendAnalyze(self, inverse=False):
        if inverse and self.board.board.opIsChecked():
            # Many engines don't like positions able to take down enemy
            # king. Therefore we just return the "kill king" move
            # automaticaly
            self.emit(
                "analyze",
                [
                    (
                        self.board.ply,
                        [toAN(self.board, getMoveKillingKing(self.board))],
                        MATE_VALUE - 1,
                        "1",
                        "",
                    )
                ],
            )
            return

        print("post", file=self.engine)
        print("analyze", file=self.engine)
        self.engineIsAnalyzing = True

        if not conf.get("infinite_depth"):
            self.__setDepth(conf.get("max_depth_spin"))
        if not conf.get("infinite_analysis"):
            loop = asyncio.get_event_loop()
            loop.call_later(conf.get("max_analysis_spin"), self.__stop_analyze)

    def __printColor(self):
        if self.features["colors"]:  # or self.mode == INVERSE_ANALYZING:
            if self.board.color == WHITE:
                print("white", file=self.engine)
            else:
                print("black", file=self.engine)

    def __setBoard(self, board):
        if self.features["setboard"]:
            self.__tellEngineToStopPlayingCurrentColor()
            fen = board.asFen(enable_bfen=False)
            if self.mode == INVERSE_ANALYZING:
                fen_arr = fen.split()
                if not self.board.board.opIsChecked():
                    if fen_arr[1] == "b":
                        fen_arr[1] = "w"
                    else:
                        fen_arr[1] = "b"
                fen = " ".join(fen_arr)
            print("setboard %s" % fen, file=self.engine)
        else:
            # "edit" doesn't support seirawan (fairymax)
            if board.variant == SCHESS:
                return
            # Kludge to set black to move, avoiding the troublesome and now
            # deprecated "black" command. - Equal to the one xboard uses
            self.__tellEngineToStopPlayingCurrentColor()
            if board.color == BLACK:
                print("a2a3", file=self.engine)
            print("edit", file=self.engine)
            print("#", file=self.engine)
            for color in WHITE, BLACK:
                for y_loc, row in enumerate(board.data):
                    for x_loc, piece in row.items():
                        if not piece or piece.color != color:
                            continue
                        sign = reprSign[piece.sign]
                        cord = repr(Cord(x_loc, y_loc))
                        print(sign + cord, file=self.engine)
                print("c", file=self.engine)
            print(".", file=self.engine)

    # Parsing

    async def parseLine(self, proc):
        while True:
            line = await wait_signal(proc, "line")
            if not line:
                break
            else:
                line = line[1]
                if line[0:1] == "#":
                    # Debug line which we shall ignore as specified in CECPv2 specs
                    continue

                #        log.debug("__parseLine: line=\"%s\"" % line.strip(), extra={"task":self.defname})
                parts = whitespaces.split(line.strip())
                if parts[0] == "pong":
                    self.lastpong = int(parts[1])
                    continue

                # Illegal Move
                if parts[0].lower().find("illegal") >= 0:
                    log.warning(
                        '__parseLine: illegal move: line="{}", board={}'.format(
                            line.strip(), self.board
                        ),
                        extra={"task": self.defname},
                    )
                    if parts[-2] == "sd" and parts[-1].isdigit():
                        print("depth", parts[-1], file=self.engine)
                    continue

                # A Move (Perhaps)
                if self.board:
                    if parts[0] == "move":
                        movestr = parts[1]
                    # Old Variation
                    elif d_plus_dot_expr.match(parts[0]) and parts[1] == "...":
                        movestr = parts[2]
                    else:
                        movestr = False

                    if movestr:
                        self.waitingForMove = False
                        self.readyForMoveNowCommand = False
                        if self.engineIsInNotPlaying:
                            # If engine was set in pause just before the engine sent its
                            # move, we ignore it. However the engine has to know that we
                            # ignored it, and thus we step it one back
                            log.info(
                                "__parseLine: Discarding engine's move: %s" % movestr,
                                extra={"task": self.defname},
                            )
                            print("undo", file=self.engine)
                            continue
                        else:
                            try:
                                move = parseAny(self.board, movestr)
                            except ParsingError:
                                self.invalid_move = movestr
                                log.info(
                                    "__parseLine: ParsingError engine move: %s %s"
                                    % (movestr, self.board),
                                    extra={"task": self.defname},
                                )
                                self.end(
                                    WHITEWON if self.board.color == BLACK else BLACKWON,
                                    WON_ADJUDICATION,
                                )
                                continue

                            if validate(self.board, move):
                                self.board = None
                                self.queue.put_nowait(move)
                                continue
                            else:
                                self.invalid_move = movestr
                                log.info(
                                    "__parseLine: can't validate engine move: %s %s"
                                    % (movestr, self.board),
                                    extra={"task": self.defname},
                                )
                                self.end(
                                    WHITEWON if self.board.color == BLACK else BLACKWON,
                                    WON_ADJUDICATION,
                                )
                                continue

                # Analyzing
                if self.engineIsInNotPlaying:
                    if parts[:4] == ["0", "0", "0", "0"]:
                        # Crafty doesn't analyze until it is out of book
                        print("book off", file=self.engine)
                        continue

                    match = anare.match(line)
                    if match:
                        depth, score, time, nodes, moves = match.groups()

                        if "mat" in score.lower() or "#" in moves:
                            # Will look either like -Mat 3 or Mat3
                            scoreval = MATE_VALUE
                            if score.startswith("-"):
                                scoreval = -scoreval
                        else:
                            scoreval = int(score)

                        nps = (
                            str(int(int(nodes) / (int(time) / 100)))
                            if int(time) > 0
                            else ""
                        )

                        mvstrs = movere.findall(moves)
                        if mvstrs:
                            self.emit(
                                "analyze",
                                [
                                    (
                                        self.board.ply,
                                        mvstrs,
                                        scoreval,
                                        depth.strip(),
                                        nps,
                                    )
                                ],
                            )

                        continue

                # Offers draw
                if parts[0:2] == ["offer", "draw"]:
                    self.emit("accept", Offer(DRAW_OFFER))
                    continue

                # Resigns
                if parts[0] == "resign" or (
                    parts[0] == "tellics" and parts[1] == "resign"
                ):  # buggy crafty
                    # Previously: if "resign" in parts,
                    # however, this is too generic, since "hint", "bk",
                    # "feature option=.." and possibly other, future CECPv2
                    # commands can validly contain the word "resign" without this
                    # being an intentional resign offer.

                    self.emit("offer", Offer(RESIGNATION))
                    continue

                # if parts[0].lower() == "error":
                #    continue

                # Tell User Error
                if parts[0] == "tellusererror":
                    # We don't want to see our stop analyzer hack as an error message
                    if "8/8/8/8/8/8/8/8" in "".join(parts[1:]):
                        continue
                    # Create a non-modal non-blocking message dialog with the error:
                    from pychess.widgets import mainwindow

                    dlg = Gtk.MessageDialog(
                        mainwindow(),
                        flags=0,
                        type=Gtk.MessageType.WARNING,
                        buttons=Gtk.ButtonsType.CLOSE,
                        message_format=None,
                    )

                    # Use the engine name if already known, otherwise the defname:
                    displayname = self.name
                    if not displayname:
                        displayname = self.defname

                    # Compose the dialog text:
                    dlg.set_markup(
                        GObject.markup_escape_text(
                            _("The engine %s reports an error:") % displayname
                        )
                        + "\n\n"
                        + GObject.markup_escape_text(" ".join(parts[1:]))
                    )

                    # handle response signal so the "Close" button works:
                    dlg.connect("response", lambda dlg, x: dlg.destroy())

                    dlg.show_all()
                    continue

                # Tell Somebody
                if parts[0][:4] == "tell" and parts[0][4:] in (
                    "others",
                    "all",
                    "ics",
                    "icsnoalias",
                ):
                    log.info(
                        "Ignoring tell {}: {}".format(parts[0][4:], " ".join(parts[1:]))
                    )
                    continue

                if "feature" in parts:
                    # Some engines send features after done=1, so we will iterate after done=1 too
                    done1 = False
                    # We skip parts before 'feature', as some engines give us lines like
                    # White (1) : feature setboard=1 analyze...e="GNU Chess 5.07" done=1
                    parts = parts[parts.index("feature") :]
                    for i, pair in enumerate(parts[1:]):
                        # As "parts" is split with no thoughs on quotes or double quotes
                        # we need to do some extra handling.

                        if pair.find("=") < 0:
                            continue
                        key, value = pair.split("=", 1)

                        if key not in self.features:
                            continue

                        if value.startswith('"') and value.endswith('"'):
                            value = value[1:-1]

                        # If our pair was unfinished, like myname="GNU, we search the
                        # rest of the pairs for a quotating mark.
                        elif value[0] == '"':
                            rest = value[1:] + " " + " ".join(parts[2 + i :])
                            j = rest.find('"')
                            if j == -1:
                                log.warning(
                                    "Missing endquotation in %s feature",
                                    extra={"task": self.defname},
                                )
                                value = rest
                            else:
                                value = rest[:j]

                        elif value.isdigit():
                            value = int(value)

                        if key in self.supported_features:
                            print("accepted %s" % key, file=self.engine)
                        else:
                            print("rejected %s" % key, file=self.engine)

                        if key == "done":
                            if value == 1:
                                done1 = True
                                continue
                            elif value == 0:
                                log.info(
                                    "Adds %d seconds timeout" % TIME_OUT_SECOND,
                                    extra={"task": self.defname},
                                )
                                # This'll buy you some more time
                                self.queue.put_nowait("not ready")
                                break

                        if key == "smp" and value == 1:
                            self.options["cores"] = {
                                "name": "cores",
                                "type": "spin",
                                "default": 1,
                                "min": 1,
                                "max": 64,
                            }
                        elif key == "memory" and value == 1:
                            self.options["memory"] = {
                                "name": "memory",
                                "type": "spin",
                                "default": 32,
                                "min": 1,
                                "max": 4096,
                            }
                        elif key == "option" and key != "done":
                            option = self.__parse_option(value)
                            self.options[option["name"]] = option
                        else:
                            self.features[key] = value

                        if key == "myname" and not self.name:
                            self.setName(value)

                    if done1:
                        # Start a new game before using the engine:
                        # (CECPv2 engines)
                        print("new", file=self.engine)

                        # We are now ready for play:
                        self.emit("readyForOptions")
                        self.emit("readyForMoves")
                        self.queue.put_nowait("ready")

                # A hack to get better names in protover 1.
                # Unfortunately it wont work for now, as we don't read any lines from
                # protover 1 engines. When should we stop?
                if self.protover == 1:
                    if self.defname[0] in "".join(parts):
                        basis = self.defname[0]
                        name = " ".join(
                            itertools.dropwhile(lambda part: basis not in part, parts)
                        )
                        self.features["myname"] = name
                        if not self.name:
                            self.setName(name)

    def __parse_option(self, option):
        if " -check " in option:
            name, value = option.split(" -check ")
            return {"type": "check", "name": name, "default": bool(int(value))}
        elif " -spin " in option:
            name, value = option.split(" -spin ")
            defv, minv, maxv = value.split()
            return {
                "type": "spin",
                "name": name,
                "default": int(defv),
                "min": int(minv),
                "max": int(maxv),
            }
        elif " -slider " in option:
            name, value = option.split(" -slider ")
            defv, minv, maxv = value.split()
            return {
                "type": "spin",
                "name": name,
                "default": int(defv),
                "min": int(minv),
                "max": int(maxv),
            }
        elif " -string " in option:
            name, value = option.split(" -string ")
            return {"type": "text", "name": name, "default": value}
        elif " -file " in option:
            name, value = option.split(" -file ")
            return {"type": "text", "name": name, "default": value}
        elif " -path " in option:
            name, value = option.split(" -path ")
            return {"type": "text", "name": name, "default": value}
        elif " -combo " in option:
            name, value = option.split(" -combo ")
            choices = list(map(str.strip, value.split("///")))
            default = ""
            for choice in choices:
                if choice.startswith("*"):
                    index = choices.index(choice)
                    default = choice[1:]
                    choices[index] = default
                    break
            return {
                "type": "combo",
                "name": name,
                "default": default,
                "choices": choices,
            }
        elif " -button" in option:
            pos = option.find(" -button")
            return {"type": "button", "name": option[:pos]}
        elif " -save" in option:
            pos = option.find(" -save")
            return {"type": "button", "name": option[:pos]}
        elif " -reset" in option:
            pos = option.find(" -reset")
            return {"type": "button", "name": option[:pos]}

    # Info

    def canAnalyze(self):
        assert self.ready, "Still waiting for done=1"
        return self.features["analyze"]

    def getAnalysisLines(self):
        return 1

    def minAnalysisLines(self):
        return 1

    def maxAnalysisLines(self):
        return 1

    def requestMultiPV(self, setting):
        return 1

    def __repr__(self):
        if self.name:
            return self.name
        return self.features["myname"]



import time
import logging

from gi.repository import GObject, GLib


class LogEmitter(GObject.GObject):
    __gsignals__ = {
        "logged": (GObject.SignalFlags.RUN_FIRST, None, (object,))
    }  # list of (str, float, str, int)

    def __init__(self):
        GObject.GObject.__init__(self)

        # We store everything in this list, so that the LogDialog, which is
        # imported a little later, will have all data ever given to Log.
        # When Dialog inits, it will set this list to None, and we will stop
        # appending data to it. Ugly? Somewhat I guess.
        self.messages = []


class GLogHandler(logging.Handler):
    def __init__(self, emitter):
        logging.Handler.__init__(self)
        self.emitter = emitter

    def emit(self, record):
        def _emit(record):
            message = self.format(record)
            if self.emitter.messages is not None:
                self.emitter.messages.append(
                    (record.task, time.time(), message, record.levelno)
                )

            self.emitter.emit(
                "logged", (record.task, time.time(), message, record.levelno)
            )

        GLib.idle_add(_emit, record)


logemitter = LogEmitter()



from pychess.Utils.const import (
    EMPTY,
    PAWN,
    KNIGHT,
    BISHOP,
    ROOK,
    QUEEN,
    KING,
    ATOMICCHESS,
    BUGHOUSECHESS,
    CRAZYHOUSECHESS,
    CAMBODIANCHESS,
    MAKRUKCHESS,
    FISCHERRANDOMCHESS,
    SITTUYINCHESS,
    WILDCASTLECHESS,
    WILDCASTLESHUFFLECHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    DROP_VARIANTS,
    BLACK,
    WHITE,
    FAN_PIECES,
    CAS_FLAGS,
    NORMALCHESS,
    PLACEMENTCHESS,
    THREECHECKCHESS,
    SETUPCHESS,
    FEN_START,
    HAWK,
    ELEPHANT,
    SCHESS,
    LIGHTBRIGADECHESS,
    chrU2Sign,
    cordDic,
    reprCord,
    reprFile,
    reprSign,
    reprSignMakruk,
    reprSignSittuyin,
    A1,
    A8,
    B1,
    B8,
    C1,
    C8,
    D1,
    D8,
    E1,
    E8,
    F1,
    F8,
    G1,
    G8,
    H1,
    H8,
    DROP,
    PROMOTIONS,
    ENPASSANT,
    B_OO,
    B_OOO,
    W_OO,
    W_OOO,
    QUEEN_CASTLE,
    KING_CASTLE,
    GATINGS,
    HAWK_GATE,
    HAWK_GATE_AT_ROOK,
    ELEPHANT_GATE_AT_ROOK,
    QUEEN_PROMOTION,
)
from pychess.Utils.repr import reprColor
from .ldata import FILE, fileBits
from .attack import isAttacked
from .bitboard import clearBit, iterBits, setBit, bitPosArray
from .PolyglotHash import (
    pieceHashes,
    epHashes,
    W_OOHash,
    W_OOOHash,
    B_OOHash,
    B_OOOHash,
    colorHash,
    holdingHash,
)

################################################################################
# FEN                                                                          #
################################################################################

# This will cause applyFen to raise an exception, if halfmove clock and fullmove
# number is not specified
STRICT_FEN = False

################################################################################
# LBoard                                                                       #
################################################################################


class LBoard:
    __hash__ = None

    ini_kings = (E1, E8)
    ini_rooks = ((A1, H1), (A8, H8))

    # Final positions of castled kings and rooks
    fin_kings = ((C1, G1), (C8, G8))
    fin_rooks = ((D1, F1), (D8, F8))

    holding = (
        {
            PAWN: 0,
            KNIGHT: 0,
            BISHOP: 0,
            ROOK: 0,
            QUEEN: 0,
            HAWK: 0,
            ELEPHANT: 0,
            KING: 0,
        },
        {
            PAWN: 0,
            KNIGHT: 0,
            BISHOP: 0,
            ROOK: 0,
            QUEEN: 0,
            HAWK: 0,
            ELEPHANT: 0,
            KING: 0,
        },
    )

    def __init__(self, variant=NORMALCHESS):
        self.variant = variant

        self.nags = []
        # children can contain comments and variations
        # variations are lists of lboard objects
        self.children = []

        # the next and prev lboard objects in the variation list
        self.next = None
        self.prev = None

        # The high level owner Board (with Piece objects) in gamemodel
        self.pieceBoard = None

        # This will True except in so called null_board
        # null_board act as parent of the variation
        # when we add a variation to last played board from hint panel
        self.fen_was_applied = False
        self.plyCount = 0

    @property
    def lastMove(self):
        return (
            self.hist_move[-1]
            if self.fen_was_applied and len(self.hist_move) > 0
            else None
        )

    def repetitionCount(self, draw_threshold=3):
        rc = 1
        for ply in range(4, 1 + min(len(self.hist_hash), self.fifty), 2):
            if self.hist_hash[-ply] == self.hash:
                rc += 1
                if rc >= draw_threshold:
                    break
        return rc

    def iniAtomic(self):
        self.hist_exploding_around = []

    def iniHouse(self):
        self.promoted = [0] * 64
        self.capture_promoting = False
        self.hist_capture_promoting = []
        self.holding = (
            {
                PAWN: 0,
                KNIGHT: 0,
                BISHOP: 0,
                ROOK: 0,
                QUEEN: 0,
                HAWK: 0,
                ELEPHANT: 0,
                KING: 0,
            },
            {
                PAWN: 0,
                KNIGHT: 0,
                BISHOP: 0,
                ROOK: 0,
                QUEEN: 0,
                HAWK: 0,
                ELEPHANT: 0,
                KING: 0,
            },
        )

    def iniCambodian(self):
        self.ini_kings = (D1, E8)
        self.ini_queens = (E1, D8)
        self.is_first_move = {KING: [True, True], QUEEN: [True, True]}
        self.hist_is_first_move = []

    def iniSchess(self):
        self.virgin = [0, 0]
        self.hist_virgin = []

    def applyFen(self, fenstr):
        """Applies the fenstring to the board.
        If the string is not properly
        written a SyntaxError will be raised, having its message ending in
        Pos(%d) specifying the string index of the problem.
        if an error is found, no changes will be made to the board."""

        assert (
            not self.fen_was_applied
        ), "The applyFen() method can be used on new LBoard objects only!"

        # Set board to empty on Black's turn (which Polyglot-hashes to 0)
        self.blocker = 0

        self.friends = [0, 0]
        self.kings = [-1, -1]

        # this variable is a 2-dimmensionnal array, each case containing a bitboard
        # self.boards[color] contains an array of bitboards, each representing the position of the pieces
        # use example : self.boards[color][KNIGHT]
        self.boards = ([0] * 9, [0] * 9)

        self.enpassant = None  # cord which can be captured by enpassant or None
        self.color = BLACK
        self.castling = 0  # The castling availability in the position
        self.hasCastled = [False, False]
        self.fifty = 0  # A ply counter for the fifty moves rule
        self.plyCount = 0

        self.checked = None
        self.opchecked = None

        self.arBoard = [0] * 64

        self.hash = 0
        self.pawnhash = 0

        #  Data from the position's history:
        self.hist_move = []  # The move that was applied to get the position
        self.hist_tpiece = []
        # The piece the move captured, == EMPTY for normal moves
        self.hist_enpassant = []
        self.hist_castling = []
        self.hist_hash = []
        self.hist_fifty = []
        self.hist_checked = []
        self.hist_opchecked = []

        # piece counts
        self.pieceCount = ([0] * 9, [0] * 9)

        # initial cords of rooks and kings for castling in Chess960
        if self.variant == FISCHERRANDOMCHESS:
            self.ini_kings = [None, None]
            self.ini_rooks = ([None, None], [None, None])

        elif self.variant in (WILDCASTLECHESS, WILDCASTLESHUFFLECHESS):
            self.ini_kings = [None, None]
            self.fin_kings = ([None, None], [None, None])
            self.fin_rooks = ([None, None], [None, None])

        elif self.variant == ATOMICCHESS:
            self.iniAtomic()

        elif self.variant == THREECHECKCHESS:
            self.remaining_checks = [3, 3]

        elif self.variant == CAMBODIANCHESS:
            self.iniCambodian()

        elif self.variant == SCHESS:
            self.iniSchess()

        if self.variant in DROP_VARIANTS:
            self.iniHouse()

            # Get information
        parts = fenstr.split()
        castChr = "-"
        epChr = "-"
        fiftyChr = "0"
        moveNoChr = "1"
        if STRICT_FEN and len(parts) != 6:
            raise SyntaxError(_("FEN needs 6 data fields. \n\n%s") % fenstr)
        elif len(parts) < 2:
            raise SyntaxError(
                _("FEN needs at least 2 data fields in fenstr. \n\n%s") % fenstr
            )
        elif len(parts) >= 7 and self.variant == THREECHECKCHESS:
            pieceChrs, colChr, castChr, epChr, checksChr, fiftyChr, moveNoChr = parts[
                :7
            ]
            self.remaining_checks = list(map(int, checksChr.split("+")))
        elif len(parts) >= 6:
            pieceChrs, colChr, castChr, epChr, fiftyChr, moveNoChr = parts[:6]
        elif len(parts) == 5:
            pieceChrs, colChr, castChr, epChr, fiftyChr = parts
        elif len(parts) == 4:
            if parts[2].isdigit() and parts[3].isdigit():
                # xboard FEN usage for asian variants
                pieceChrs, colChr, fiftyChr, moveNoChr = parts
            else:
                pieceChrs, colChr, castChr, epChr = parts
        elif len(parts) == 3:
            pieceChrs, colChr, castChr = parts
        else:
            pieceChrs, colChr = parts

        slashes = pieceChrs.count("/")
        if slashes < 7:
            raise SyntaxError(
                _("Needs 7 slashes in piece placement field. \n\n%s") % fenstr
            )

        if not colChr.lower() in ("w", "b"):
            raise SyntaxError(
                _("Active color field must be one of w or b. \n\n%s") % fenstr
            )

        if castChr != "-":
            for Chr in castChr:
                if self.variant == FISCHERRANDOMCHESS or self.variant == SETUPCHESS:
                    valid_chars = "ABCDEFGHKQ"
                elif self.variant == CAMBODIANCHESS:
                    valid_chars = "DE"
                elif self.variant == SCHESS:
                    valid_chars = "KQABCDEFGH"
                elif self.variant == LIGHTBRIGADECHESS:
                    valid_chars = ""
                else:
                    valid_chars = "KQ"

                if Chr.upper() not in valid_chars:
                    raise SyntaxError(
                        _("Castling availability field is not legal. \n\n%s") % fenstr
                    )

        if epChr != "-" and epChr not in cordDic:
            raise SyntaxError(_("En passant cord is not legal. \n\n%s") % fenstr)

        # Parse piece placement field
        promoted = False
        # if there is a holding within [] we change it to BFEN style first
        if pieceChrs.endswith("]"):
            pieceChrs = pieceChrs[:-1].replace("[", "/").replace("-", "")
        for r, rank in enumerate(pieceChrs.split("/")):
            cord = (7 - r) * 8
            for char in rank:
                if r > 7:
                    # After the 8.rank BFEN can contain holdings (captured pieces)
                    # "~" after a piece letter denotes promoted piece
                    if r == 8 and self.variant in DROP_VARIANTS:
                        color = char.islower() and BLACK or WHITE
                        piece = chrU2Sign[char.upper()]
                        self.holding[color][piece] += 1
                        self.hash ^= holdingHash[color][piece][
                            self.holding[color][piece]
                        ]
                        continue
                    else:
                        break

                if char.isdigit():
                    cord += int(char)
                elif char == "~":
                    promoted = True
                else:
                    color = char.islower() and BLACK or WHITE
                    piece = chrU2Sign[char.upper()]
                    self._addPiece(cord, piece, color)
                    self.pieceCount[color][piece] += 1

                    if self.variant in DROP_VARIANTS and promoted:
                        self.promoted[cord] = 1
                        promoted = False

                    if self.variant == CAMBODIANCHESS:
                        if piece == KING and self.kings[color] != self.ini_kings[color]:
                            self.is_first_move[KING][color] = False
                        if piece == QUEEN and cord != self.ini_queens[color]:
                            self.is_first_move[QUEEN][color] = False

                    cord += 1

            if self.variant == FISCHERRANDOMCHESS:
                # Save ranks fo find outermost rooks
                # if KkQq was used in castling rights
                if r == 0:
                    rank8 = rank
                elif r == 7:
                    rank1 = rank

        # Parse active color field

        if colChr.lower() == "w":
            self.setColor(WHITE)
        else:
            self.setColor(BLACK)

        # Parse castling availability
        # TODO: S-Chess gating

        castling = 0
        for char in castChr:
            if self.variant == FISCHERRANDOMCHESS:
                if char in reprFile:
                    if char < reprCord[self.kings[BLACK]][0]:
                        castling |= B_OOO
                        self.ini_rooks[1][0] = reprFile.index(char) + 56
                    else:
                        castling |= B_OO
                        self.ini_rooks[1][1] = reprFile.index(char) + 56
                elif char in [c.upper() for c in reprFile]:
                    if char < reprCord[self.kings[WHITE]][0].upper():
                        castling |= W_OOO
                        self.ini_rooks[0][0] = reprFile.index(char.lower())
                    else:
                        castling |= W_OO
                        self.ini_rooks[0][1] = reprFile.index(char.lower())
                elif char == "K":
                    castling |= W_OO
                    self.ini_rooks[0][1] = rank1.rfind("R")
                elif char == "Q":
                    castling |= W_OOO
                    self.ini_rooks[0][0] = rank1.find("R")
                elif char == "k":
                    castling |= B_OO
                    self.ini_rooks[1][1] = rank8.rfind("r") + 56
                elif char == "q":
                    castling |= B_OOO
                    self.ini_rooks[1][0] = rank8.find("r") + 56
            else:
                if char == "K":
                    castling |= W_OO
                elif char == "Q":
                    castling |= W_OOO
                elif char == "k":
                    castling |= B_OO
                elif char == "q":
                    castling |= B_OOO

            if self.variant == SCHESS:
                if char == "K":
                    self.virgin[WHITE] = setBit(self.virgin[WHITE], E1)
                    self.virgin[WHITE] = setBit(self.virgin[WHITE], H1)
                elif char == "Q":
                    self.virgin[WHITE] = setBit(self.virgin[WHITE], A1)
                    self.virgin[WHITE] = setBit(self.virgin[WHITE], E1)
                elif char == "k":
                    self.virgin[BLACK] = setBit(self.virgin[BLACK], E8)
                    self.virgin[BLACK] = setBit(self.virgin[BLACK], H8)
                elif char == "q":
                    self.virgin[BLACK] = setBit(self.virgin[BLACK], A8)
                    self.virgin[BLACK] = setBit(self.virgin[BLACK], E8)
                else:
                    side = WHITE if char.isupper() else BLACK
                    cord = ord(char) - 65 if side == WHITE else ord(char) - 97 + 56
                    self.virgin[side] = setBit(self.virgin[side], cord)

        if self.variant in (
            WILDCASTLECHESS,
            WILDCASTLESHUFFLECHESS,
            FISCHERRANDOMCHESS,
        ):
            self.ini_kings[WHITE] = self.kings[WHITE]
            self.ini_kings[BLACK] = self.kings[BLACK]
            if self.variant in (WILDCASTLECHESS, WILDCASTLESHUFFLECHESS):
                if self.ini_kings[WHITE] == D1 and self.ini_kings[BLACK] == D8:
                    self.fin_kings = ([B1, F1], [B8, F8])
                    self.fin_rooks = ([C1, E1], [C8, E8])
                elif self.ini_kings[WHITE] == D1:
                    self.fin_kings = ([B1, F1], [C8, G8])
                    self.fin_rooks = ([C1, E1], [D8, F8])
                elif self.ini_kings[BLACK] == D8:
                    self.fin_kings = ([C1, G1], [B8, F8])
                    self.fin_rooks = ([D1, F1], [C8, E8])
                else:
                    self.fin_kings = ([C1, G1], [C8, G8])
                    self.fin_rooks = ([D1, F1], [D8, F8])

        self.setCastling(castling)

        # Parse en passant target sqaure

        if epChr == "-":
            self.setEnpassant(None)
        else:
            self.setEnpassant(cordDic[epChr])

        # Parse halfmove clock field

        if fiftyChr.isdigit():
            self.fifty = int(fiftyChr)
        else:
            self.fifty = 0

        # Parse fullmove number

        if moveNoChr.isdigit():
            movenumber = max(int(moveNoChr), 1) * 2 - 2
            if self.color == BLACK:
                movenumber += 1
            self.plyCount = movenumber
        else:
            self.plyCount = 1

        self.fen_was_applied = True

    def isChecked(self):
        if self.variant == SUICIDECHESS or self.variant == GIVEAWAYCHESS:
            return False
        elif self.variant == ATOMICCHESS:
            if not self.boards[self.color][KING]:
                return False
            if (
                -2 < (self.kings[0] >> 3) - (self.kings[1] >> 3) < 2
                and -2 < (self.kings[0] & 7) - (self.kings[1] & 7) < 2
            ):
                return False
        if self.checked is None:
            kingcord = self.kings[self.color]
            if kingcord == -1:
                return False
            self.checked = isAttacked(self, kingcord, 1 - self.color, ischecked=True)
        return self.checked

    def opIsChecked(self):
        if self.variant == SUICIDECHESS or self.variant == GIVEAWAYCHESS:
            return False
        elif self.variant == ATOMICCHESS:
            if not self.boards[1 - self.color][KING]:
                return False
            if (
                -2 < (self.kings[0] >> 3) - (self.kings[1] >> 3) < 2
                and -2 < (self.kings[0] & 7) - (self.kings[1] & 7) < 2
            ):
                return False
        if self.opchecked is None:
            kingcord = self.kings[1 - self.color]
            if kingcord == -1:
                return False
            self.opchecked = isAttacked(self, kingcord, self.color, ischecked=True)
        return self.opchecked

    def willLeaveInCheck(self, move):
        if self.variant == SUICIDECHESS or self.variant == GIVEAWAYCHESS:
            return False
        board_clone = self.clone()
        board_clone.applyMove(move)
        return board_clone.opIsChecked()

    def willGiveCheck(self, move):
        board_clone = self.clone()
        board_clone.applyMove(move)
        return board_clone.isChecked()

    def _addPiece(self, cord, piece, color):
        _setBit = setBit
        self.boards[color][piece] = _setBit(self.boards[color][piece], cord)
        self.friends[color] = _setBit(self.friends[color], cord)
        self.blocker = _setBit(self.blocker, cord)

        if piece == PAWN:
            self.pawnhash ^= pieceHashes[color][PAWN][cord]
        elif piece == KING:
            self.kings[color] = cord
        self.hash ^= pieceHashes[color][piece][cord]
        self.arBoard[cord] = piece

    def _removePiece(self, cord, piece, color):
        _clearBit = clearBit
        self.boards[color][piece] = _clearBit(self.boards[color][piece], cord)
        self.friends[color] = _clearBit(self.friends[color], cord)
        self.blocker = _clearBit(self.blocker, cord)

        if piece == PAWN:
            self.pawnhash ^= pieceHashes[color][PAWN][cord]

        self.hash ^= pieceHashes[color][piece][cord]
        self.arBoard[cord] = EMPTY

    def setColor(self, color):
        if color == self.color:
            return
        self.color = color
        self.hash ^= colorHash

    def setCastling(self, castling):
        if self.castling == castling:
            return

        if castling & W_OO != self.castling & W_OO:
            self.hash ^= W_OOHash
        if castling & W_OOO != self.castling & W_OOO:
            self.hash ^= W_OOOHash
        if castling & B_OO != self.castling & B_OO:
            self.hash ^= B_OOHash
        if castling & B_OOO != self.castling & B_OOO:
            self.hash ^= B_OOOHash

        self.castling = castling

    def setEnpassant(self, epcord):
        # Strip the square if there's no adjacent enemy pawn to make the capture
        if epcord is not None:
            sideToMove = epcord >> 3 == 2 and BLACK or WHITE
            fwdPawns = self.boards[sideToMove][PAWN]
            if sideToMove == WHITE:
                fwdPawns >>= 8
            else:
                fwdPawns <<= 8
            pawnTargets = (fwdPawns & ~fileBits[0]) << 1
            pawnTargets |= (fwdPawns & ~fileBits[7]) >> 1
            if not pawnTargets & bitPosArray[epcord]:
                epcord = None

        if self.enpassant == epcord:
            return
        if self.enpassant is not None:
            self.hash ^= epHashes[self.enpassant & 7]
        if epcord is not None:
            self.hash ^= epHashes[epcord & 7]
        self.enpassant = epcord

    # @profile
    def applyMove(self, move):
        flag = move >> 12

        fcord = (move >> 6) & 63
        tcord = move & 63

        fpiece = fcord if flag == DROP else self.arBoard[fcord]
        tpiece = self.arBoard[tcord]

        color = self.color
        opcolor = 1 - self.color
        castling = self.castling

        kcastle = flag == KING_CASTLE or (
            self.variant == SCHESS
            and (
                (fpiece == KING and fcord - tcord == -2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and fcord - tcord > 0
                )
            )
        )
        qcastle = flag == QUEEN_CASTLE or (
            self.variant == SCHESS
            and (
                (fpiece == KING and fcord - tcord == 2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and fcord - tcord < 0
                )
            )
        )

        self.hist_move.append(move)
        self.hist_enpassant.append(self.enpassant)
        self.hist_castling.append(self.castling)
        self.hist_hash.append(self.hash)
        self.hist_fifty.append(self.fifty)
        self.hist_checked.append(self.checked)
        self.hist_opchecked.append(self.opchecked)
        if self.variant in DROP_VARIANTS and self.variant != SCHESS:
            self.hist_capture_promoting.append(self.capture_promoting)
        if self.variant == CAMBODIANCHESS:
            self.hist_is_first_move.append(
                {KING: self.is_first_move[KING][:], QUEEN: self.is_first_move[QUEEN][:]}
            )
        elif self.variant == SCHESS:
            self.hist_virgin.append(self.virgin[:])

        self.opchecked = None
        self.checked = None

        # null move
        if fcord == tcord and flag != DROP and flag != QUEEN_PROMOTION:
            self.setColor(opcolor)
            self.plyCount += 1
            return move

        if self.variant == CAMBODIANCHESS:
            if fpiece == KING and self.is_first_move[KING][color]:
                self.is_first_move[KING][color] = False
            elif fpiece == QUEEN and self.is_first_move[QUEEN][color]:
                self.is_first_move[QUEEN][color] = False
        elif self.variant == SCHESS:
            if qcastle:
                self.virgin[color] = clearBit(self.virgin[color], self.ini_kings[color])
                self.virgin[color] = clearBit(
                    self.virgin[color], self.ini_rooks[color][0]
                )
            elif kcastle:
                self.virgin[color] = clearBit(self.virgin[color], self.ini_kings[color])
                self.virgin[color] = clearBit(
                    self.virgin[color], self.ini_rooks[color][1]
                )
            elif fcord in iterBits(self.virgin[color]):
                self.virgin[color] = clearBit(self.virgin[color], fcord)

        # Castling moves can be represented strangely, so normalize them.
        if kcastle or qcastle:
            side = 0 if qcastle else 1
            fpiece = KING
            tpiece = EMPTY  # In FRC, there may be a rook there, but the king doesn't capture it.
            gcord = fcord  # save gating cord
            fcord = self.ini_kings[color]
            if FILE(fcord) == 3 and self.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                side = 0 if side == 1 else 1
            tcord = self.fin_kings[color][side]
            rookf = self.ini_rooks[color][side]
            rookt = self.fin_rooks[color][side]

        # Capture (sittuyin in place promotion is not capture move!)
        if tpiece != EMPTY and fcord != tcord:
            self._removePiece(tcord, tpiece, opcolor)
            self.pieceCount[opcolor][tpiece] -= 1
            if self.variant in DROP_VARIANTS:
                if self.promoted[tcord]:
                    if self.variant == CRAZYHOUSECHESS:
                        self.holding[color][PAWN] += 1
                        self.hash ^= holdingHash[color][PAWN][self.holding[color][PAWN]]
                    self.capture_promoting = True
                else:
                    if self.variant == CRAZYHOUSECHESS:
                        self.holding[color][tpiece] += 1
                        self.hash ^= holdingHash[color][tpiece][
                            self.holding[color][tpiece]
                        ]
                    self.capture_promoting = False
            elif self.variant == ATOMICCHESS:
                from pychess.Variants.atomic import piecesAround

                apieces = [
                    (fcord, fpiece, color),
                ]
                for acord, apiece, acolor in piecesAround(self, tcord):
                    if apiece != PAWN and acord != fcord:
                        self._removePiece(acord, apiece, acolor)
                        self.pieceCount[acolor][apiece] -= 1
                        apieces.append((acord, apiece, acolor))
                    if apiece == ROOK and acord != fcord:
                        if acord == self.ini_rooks[opcolor][0]:
                            castling &= ~CAS_FLAGS[opcolor][0]
                        elif acord == self.ini_rooks[opcolor][1]:
                            castling &= ~CAS_FLAGS[opcolor][1]
                self.hist_exploding_around.append(apieces)
        self.hist_tpiece.append(tpiece)

        # Remove moving piece(s), then add them at their destination.
        if flag == DROP:
            if self.variant in DROP_VARIANTS:
                assert self.holding[color][fpiece] > 0
            self.holding[color][fpiece] -= 1
            self.hash ^= holdingHash[color][fpiece][self.holding[color][fpiece]]
            self.pieceCount[color][fpiece] += 1
        else:
            self._removePiece(fcord, fpiece, color)

        if kcastle or qcastle:
            self._removePiece(rookf, ROOK, color)
            self._addPiece(rookt, ROOK, color)
            self.hasCastled[color] = True

        if flag == ENPASSANT:
            takenPawnC = tcord + (color == WHITE and -8 or 8)
            self._removePiece(takenPawnC, PAWN, opcolor)
            self.pieceCount[opcolor][PAWN] -= 1
            if self.variant == CRAZYHOUSECHESS:
                self.holding[color][PAWN] += 1
                self.hash ^= holdingHash[color][PAWN][self.holding[color][PAWN]]
            elif self.variant == ATOMICCHESS:
                from pychess.Variants.atomic import piecesAround

                apieces = [
                    (fcord, fpiece, color),
                ]
                for acord, apiece, acolor in piecesAround(self, tcord):
                    if apiece != PAWN and acord != fcord:
                        self._removePiece(acord, apiece, acolor)
                        self.pieceCount[acolor][apiece] -= 1
                        apieces.append((acord, apiece, acolor))
                self.hist_exploding_around.append(apieces)
        elif flag in PROMOTIONS:
            # Pretend the pawn changes into a piece before reaching its destination.
            fpiece = flag - 2
            self.pieceCount[color][fpiece] += 1
            self.pieceCount[color][PAWN] -= 1

        if self.variant in DROP_VARIANTS:
            if tpiece == EMPTY:
                self.capture_promoting = False

            if flag in PROMOTIONS:
                self.promoted[tcord] = 1
            elif flag != DROP:
                if self.promoted[fcord]:
                    self.promoted[fcord] = 0
                    self.promoted[tcord] = 1
                elif tpiece != EMPTY:
                    self.promoted[tcord] = 0

        if flag in GATINGS:
            gpiece = HAWK if flag in (HAWK_GATE, HAWK_GATE_AT_ROOK) else ELEPHANT
            self.holding[color][gpiece] -= 1
            self.hash ^= holdingHash[color][gpiece][self.holding[color][gpiece]]
            self.pieceCount[color][gpiece] += 1
            self._addPiece(gcord if (kcastle or qcastle) else fcord, gpiece, color)

        if self.variant == ATOMICCHESS and (tpiece != EMPTY or flag == ENPASSANT):
            self.pieceCount[color][fpiece] -= 1
        else:
            self._addPiece(tcord, fpiece, color)

        if fpiece == PAWN and abs(fcord - tcord) == 16:
            self.setEnpassant((fcord + tcord) // 2)
        else:
            self.setEnpassant(None)

        if tpiece == EMPTY and fpiece != PAWN:
            self.fifty += 1
        else:
            self.fifty = 0

        # Clear castle flags
        king = self.ini_kings[color]
        wildcastle = FILE(king) == 3 and self.variant in (
            WILDCASTLECHESS,
            WILDCASTLESHUFFLECHESS,
        )
        if fpiece == KING:
            castling &= ~CAS_FLAGS[color][0]
            castling &= ~CAS_FLAGS[color][1]
        elif fpiece == ROOK:
            if fcord == self.ini_rooks[color][0]:
                side = 1 if wildcastle else 0
                castling &= ~CAS_FLAGS[color][side]
            elif fcord == self.ini_rooks[color][1]:
                side = 0 if wildcastle else 1
                castling &= ~CAS_FLAGS[color][side]
        if tpiece == ROOK:
            if tcord == self.ini_rooks[opcolor][0]:
                side = 1 if wildcastle else 0
                castling &= ~CAS_FLAGS[opcolor][side]
            elif tcord == self.ini_rooks[opcolor][1]:
                side = 0 if wildcastle else 1
                castling &= ~CAS_FLAGS[opcolor][side]

        if self.variant == PLACEMENTCHESS and self.plyCount == 15:
            castling = 0
            if self.arBoard[A1] == ROOK and self.arBoard[E1] == KING:
                castling |= W_OOO
            if self.arBoard[H1] == ROOK and self.arBoard[E1] == KING:
                castling |= W_OO
            if self.arBoard[A8] == ROOK and self.arBoard[E8] == KING:
                castling |= B_OOO
            if self.arBoard[H8] == ROOK and self.arBoard[E8] == KING:
                castling |= B_OO

        self.setCastling(castling)

        self.setColor(opcolor)
        self.plyCount += 1

    def popMove(self):
        # Note that we remove the last made move, which was not made by boards
        # current color, but by its opponent
        color = 1 - self.color
        opcolor = self.color

        move = self.hist_move.pop()
        cpiece = self.hist_tpiece.pop()

        flag = move >> 12

        fcord = (move >> 6) & 63
        tcord = move & 63
        tpiece = self.arBoard[tcord]

        if fcord == tcord and flag != DROP and flag != QUEEN_PROMOTION:
            self.setColor(color)
            return

        if flag in GATINGS:
            gpiece = HAWK if flag in (HAWK_GATE, HAWK_GATE_AT_ROOK) else ELEPHANT
            self.holding[color][gpiece] += 1
            self.hash ^= holdingHash[color][gpiece][self.holding[color][gpiece]]
            self.pieceCount[color][gpiece] -= 1
            self._removePiece(fcord, gpiece, color)

        kcastle = flag == KING_CASTLE or (
            self.variant == SCHESS
            and (
                (tpiece == KING and fcord - tcord == -2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and fcord - tcord > 0
                )
            )
        )
        qcastle = flag == QUEEN_CASTLE or (
            self.variant == SCHESS
            and (
                (tpiece == KING and fcord - tcord == 2)
                or (
                    flag in (HAWK_GATE_AT_ROOK, ELEPHANT_GATE_AT_ROOK)
                    and fcord - tcord < 0
                )
            )
        )

        # Castling moves can be represented strangely, so normalize them.
        if kcastle or qcastle:
            side = 0 if qcastle else 1
            tpiece = KING
            fcord = self.ini_kings[color]
            if FILE(fcord) == 3 and self.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            ):
                side = 0 if side == 1 else 1
            tcord = self.fin_kings[color][side]
            rookf = self.ini_rooks[color][side]
            rookt = self.fin_rooks[color][side]
            self._removePiece(tcord, tpiece, color)
            self._removePiece(rookt, ROOK, color)
            self._addPiece(rookf, ROOK, color)
            self.hasCastled[color] = False
        else:
            self._removePiece(tcord, tpiece, color)

        # Put back captured piece
        if cpiece != EMPTY and fcord != tcord:
            self._addPiece(tcord, cpiece, opcolor)
            self.pieceCount[opcolor][cpiece] += 1
            if self.variant == CRAZYHOUSECHESS:
                if self.capture_promoting:
                    assert self.holding[color][PAWN] > 0
                    self.holding[color][PAWN] -= 1
                    self.hash ^= holdingHash[color][PAWN][self.holding[color][PAWN]]
                else:
                    assert self.holding[color][cpiece] > 0
                    self.holding[color][cpiece] -= 1
                    self.hash ^= holdingHash[color][cpiece][self.holding[color][cpiece]]
            elif self.variant == ATOMICCHESS:
                apieces = self.hist_exploding_around.pop()
                for acord, apiece, acolor in apieces:
                    self._addPiece(acord, apiece, acolor)
                    self.pieceCount[acolor][apiece] += 1

                    # Put back piece captured by enpassant
        if flag == ENPASSANT:
            epcord = color == WHITE and tcord - 8 or tcord + 8
            self._addPiece(epcord, PAWN, opcolor)
            self.pieceCount[opcolor][PAWN] += 1
            if self.variant == CRAZYHOUSECHESS:
                assert self.holding[color][PAWN] > 0
                self.holding[color][PAWN] -= 1
                self.hash ^= holdingHash[color][PAWN][self.holding[color][PAWN]]
            elif self.variant == ATOMICCHESS:
                apieces = self.hist_exploding_around.pop()
                for acord, apiece, acolor in apieces:
                    self._addPiece(acord, apiece, acolor)
                    self.pieceCount[acolor][apiece] += 1

            # Un-promote pawn
        if flag in PROMOTIONS:
            tpiece = PAWN
            self.pieceCount[color][flag - 2] -= 1
            self.pieceCount[color][PAWN] += 1

        # Put back moved piece
        if flag == DROP:
            try:
                self.holding[color][tpiece] += 1
            except KeyError:
                print(move, color, tpiece, self.holding)
                print(self)
            self.hash ^= holdingHash[color][tpiece][self.holding[color][tpiece]]
            self.pieceCount[color][tpiece] -= 1
        else:
            if not (
                self.variant == ATOMICCHESS and (cpiece != EMPTY or flag == ENPASSANT)
            ):
                self._addPiece(fcord, tpiece, color)

        if self.variant in DROP_VARIANTS and self.variant != SCHESS:
            if flag != DROP:
                if self.promoted[tcord] and (flag not in PROMOTIONS):
                    self.promoted[fcord] = 1
                if self.capture_promoting:
                    self.promoted[tcord] = 1
                else:
                    self.promoted[tcord] = 0
            self.capture_promoting = self.hist_capture_promoting.pop()

        if self.variant == CAMBODIANCHESS:
            self.is_first_move = self.hist_is_first_move.pop()
        elif self.variant == SCHESS:
            self.virgin = self.hist_virgin.pop()

        self.setColor(color)

        self.checked = self.hist_checked.pop()
        self.opchecked = self.hist_opchecked.pop()
        self.enpassant = self.hist_enpassant.pop()
        self.castling = self.hist_castling.pop()
        self.hash = self.hist_hash.pop()
        self.fifty = self.hist_fifty.pop()
        self.plyCount -= 1

    def __eq__(self, other):
        if not (
            other is not None
            and self.fen_was_applied
            and other.fen_was_applied
            and self.hash == other.hash
            and self.plyCount == other.plyCount
        ):
            return False

        b0, b1 = self.prev, other.prev
        ok = True
        while ok and b0 is not None and b1 is not None:
            if not (
                b0.fen_was_applied
                and b1.fen_was_applied
                and b0.hash == b1.hash
                and b0.plyCount == b1.plyCount
            ):
                ok = False
            else:
                b0, b1 = b0.prev, b1.prev
        return ok

    def __ne__(self, other):
        return not self.__eq__(other)

    def reprCastling(self):
        if not self.castling:
            return "-"
        else:
            strs = []
            if self.variant == FISCHERRANDOMCHESS:
                if self.castling & W_OO:
                    strs.append(reprCord[self.ini_rooks[0][1]][0].upper())
                if self.castling & W_OOO:
                    strs.append(reprCord[self.ini_rooks[0][0]][0].upper())
                if self.castling & B_OO:
                    strs.append(reprCord[self.ini_rooks[1][1]][0])
                if self.castling & B_OOO:
                    strs.append(reprCord[self.ini_rooks[1][0]][0])
            else:
                if self.castling & W_OO:
                    strs.append("K")
                if self.castling & W_OOO:
                    strs.append("Q")
                if self.castling & B_OO:
                    strs.append("k")
                if self.castling & B_OOO:
                    strs.append("q")
            return "".join(strs)

    def prepr(self, ascii=False):
        if not self.fen_was_applied:
            return "LBoard without applied FEN"
        b = "#" + reprColor[self.color] + " "
        b += self.reprCastling() + " "
        b += self.enpassant is not None and reprCord[self.enpassant] or "-"
        b += "\n# "
        rows = [self.arBoard[i : i + 8] for i in range(0, 64, 8)][::-1]
        for r, row in enumerate(rows):
            for i, piece in enumerate(row):
                if piece != EMPTY:
                    if bitPosArray[(7 - r) * 8 + i] & self.friends[WHITE]:
                        assert self.boards[WHITE][
                            piece
                        ], "self.boards doesn't match self.arBoard !!!"
                        sign = reprSign[piece] if ascii else FAN_PIECES[WHITE][piece]
                    else:
                        assert self.boards[BLACK][
                            piece
                        ], "self.boards doesn't match self.arBoard !!!"
                        sign = (
                            reprSign[piece].lower()
                            if ascii
                            else FAN_PIECES[BLACK][piece]
                        )
                    b += sign
                else:
                    b += "."
                b += " "
            b += "\n# "

        if self.variant in DROP_VARIANTS:
            for color in (BLACK, WHITE):
                holding = self.holding[color]
                b += "\n# [%s]" % "".join(
                    [
                        (
                            reprSign[piece]
                            if ascii
                            else FAN_PIECES[color][piece] * holding[piece]
                        )
                        for piece in holding
                        if holding[piece] > 0
                    ]
                )
        return b

    def __repr__(self):
        return self.prepr()

    def asFen(self, enable_bfen=False):
        fenstr = []

        rows = [self.arBoard[i : i + 8] for i in range(0, 64, 8)][::-1]
        for r, row in enumerate(rows):
            empty = 0
            for i, piece in enumerate(row):
                if piece != EMPTY:
                    if empty > 0:
                        fenstr.append(str(empty))
                        empty = 0
                    if self.variant in (CAMBODIANCHESS, MAKRUKCHESS):
                        sign = reprSignMakruk[piece]
                    elif self.variant == SITTUYINCHESS:
                        sign = reprSignSittuyin[piece]
                    else:
                        sign = reprSign[piece]
                    if bitPosArray[(7 - r) * 8 + i] & self.friends[WHITE]:
                        sign = sign.upper()
                    else:
                        sign = sign.lower()
                    fenstr.append(sign)
                    if self.variant in (BUGHOUSECHESS, CRAZYHOUSECHESS):
                        if self.promoted[r * 8 + i]:
                            fenstr.append("~")
                else:
                    empty += 1
            if empty > 0:
                fenstr.append(str(empty))
            if r != 7:
                fenstr.append("/")

        if self.variant in DROP_VARIANTS:
            holding_pieces = []
            for color in (BLACK, WHITE):
                holding = self.holding[color]
                for piece in holding:
                    if holding[piece] > 0:
                        if self.variant == SITTUYINCHESS:
                            sign = reprSignSittuyin[piece]
                        else:
                            sign = reprSign[piece]
                        sign = sign.upper() if color == WHITE else sign.lower()
                        holding_pieces.append(sign * holding[piece])
            if holding_pieces:
                if enable_bfen:
                    fenstr.append("/")
                    fenstr += holding_pieces
                else:
                    fenstr.append("[")
                    fenstr += holding_pieces
                    fenstr.append("]")

        fenstr.append(" ")

        fenstr.append(self.color == WHITE and "w" or "b")
        fenstr.append(" ")

        if self.variant == CAMBODIANCHESS:
            cast = ""
            if self.is_first_move[KING][WHITE]:
                cast += "D"
            if self.is_first_move[QUEEN][WHITE]:
                cast += "E"
            if self.is_first_move[KING][BLACK]:
                cast += "d"
            if self.is_first_move[QUEEN][BLACK]:
                cast += "e"
            if not cast:
                cast = "-"
            fenstr.append(cast)
        elif self.variant == SCHESS:
            # TODO: remove redundant letters
            cast = ""
            if self.castling & W_OO:
                cast += "K"
            if self.castling & W_OOO:
                cast += "Q"
            if self.holding[WHITE][HAWK] > 0 or self.holding[WHITE][ELEPHANT] > 0:
                for bit in iterBits(self.virgin[WHITE]):
                    cast += "ABCDEFGH"[bit]
            if self.castling & B_OO:
                cast += "k"
            if self.castling & B_OOO:
                cast += "q"
            if self.holding[BLACK][HAWK] > 0 or self.holding[BLACK][ELEPHANT] > 0:
                for bit in iterBits(self.virgin[BLACK]):
                    cast += "abcdefgh"[bit - 56]
            if not cast:
                cast = "-"
            fenstr.append(cast)
        else:
            fenstr.append(self.reprCastling())
        fenstr.append(" ")

        if not self.enpassant:
            fenstr.append("-")
        else:
            fenstr.append(reprCord[self.enpassant])
        fenstr.append(" ")

        fenstr.append(str(self.fifty))
        fenstr.append(" ")

        fullmove = (self.plyCount) // 2 + 1
        fenstr.append(str(fullmove))

        return "".join(fenstr)

    def clone(self):
        copy = LBoard(self.variant)
        copy.blocker = self.blocker

        copy.friends = self.friends[:]
        copy.kings = self.kings[:]
        copy.boards = (self.boards[WHITE][:], self.boards[BLACK][:])
        copy.arBoard = self.arBoard[:]
        copy.pieceCount = (self.pieceCount[WHITE][:], self.pieceCount[BLACK][:])

        copy.color = self.color
        copy.plyCount = self.plyCount
        copy.hasCastled = self.hasCastled[:]

        copy.enpassant = self.enpassant
        copy.castling = self.castling
        copy.hash = self.hash
        copy.pawnhash = self.pawnhash
        copy.fifty = self.fifty
        copy.checked = self.checked
        copy.opchecked = self.opchecked

        copy.hist_move = self.hist_move[:]
        copy.hist_tpiece = self.hist_tpiece[:]
        copy.hist_enpassant = self.hist_enpassant[:]
        copy.hist_castling = self.hist_castling[:]
        copy.hist_hash = self.hist_hash[:]
        copy.hist_fifty = self.hist_fifty[:]
        copy.hist_checked = self.hist_checked[:]
        copy.hist_opchecked = self.hist_opchecked[:]

        if self.variant == FISCHERRANDOMCHESS:
            copy.ini_kings = self.ini_kings[:]
            copy.ini_rooks = (self.ini_rooks[0][:], self.ini_rooks[1][:])
        elif self.variant in (WILDCASTLECHESS, WILDCASTLESHUFFLECHESS):
            copy.ini_kings = self.ini_kings[:]
            copy.fin_kings = (self.fin_kings[0][:], self.fin_kings[1][:])
            copy.fin_rooks = (self.fin_rooks[0][:], self.fin_rooks[1][:])
        elif self.variant in DROP_VARIANTS:
            copy.promoted = self.promoted[:]
            copy.holding = (self.holding[0].copy(), self.holding[1].copy())
            copy.capture_promoting = self.capture_promoting
            copy.hist_capture_promoting = self.hist_capture_promoting[:]
            if self.variant == SCHESS:
                copy.virgin = self.virgin[:]
                copy.hist_virgin = self.hist_virgin[:]
        elif self.variant == ATOMICCHESS:
            copy.hist_exploding_around = [a[:] for a in self.hist_exploding_around]
        elif self.variant == THREECHECKCHESS:
            copy.remaining_checks = self.remaining_checks[:]
        elif self.variant == CAMBODIANCHESS:
            copy.ini_kings = self.ini_kings
            copy.ini_queens = self.ini_queens
            copy.is_first_move = {
                KING: self.is_first_move[KING][:],
                QUEEN: self.is_first_move[QUEEN][:],
            }
            copy.hist_is_first_move = self.hist_is_first_move[:]

        copy.fen_was_applied = self.fen_was_applied
        return copy


START_BOARD = LBoard()
START_BOARD.applyFen(FEN_START)



import os.path
import json

from gi.repository import GLib, Gtk

from pychess import VERSION
from pychess.widgets import mainwindow
from pychess.System import download_file_async, prefix

URL = "https://api.github.com/repos/pychess/pychess/releases/latest"
LINK = "https://github.com/pychess/pychess/releases"


def isgit():
    return os.path.isdir(prefix.addDataPrefix(".git"))


async def checkversion():
    if isgit():
        return

    new_version = None

    filename = await download_file_async(URL)

    if filename is not None:
        with open(filename, encoding="utf-8") as f:
            new_version = json.loads(f.read())["name"]

    def notify(new_version):
        msg_dialog = Gtk.MessageDialog(
            mainwindow(), type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK
        )

        msg = _("<b>New version %s is available!</b>" % new_version)
        msg_dialog.set_markup(msg)
        msg_dialog.format_secondary_markup(f'<a href="{LINK}">{LINK}</a>')

        msg_dialog.connect("response", lambda msg_dialog, a: msg_dialog.hide())
        msg_dialog.show()

    if new_version is not None and VERSION.split(".") < new_version.split("."):
        GLib.idle_add(notify, new_version)
